r"""
BGG Category O Dual Modules

AUTHORS:

- Travis Scrimshaw (2024-01-07): Initial version
"""

#*****************************************************************************
#       Copyright (C) 2024 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from typing import Self

from sage.algebras.lie_algebras.verma_module import ModulePrinting
from sage.categories.enumerated_sets import EnumeratedSets
from sage.categories.monoids import Monoids
from sage.combinat.free_module import CombinatorialFreeModule
from sage.data_structures.blas_dict import iaxpy
from sage.matrix.constructor import matrix
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.monoids.indexed_free_monoid import IndexedFreeAbelianMonoid, IndexedMonoid
from sage.rings.integer_ring import ZZ
from sage.sets.family import Family
from sage.sets.finite_enumerated_set import FiniteEnumeratedSet
from sage.structure.indexed_generators import IndexedGenerators
from sage.structure.parent import Parent


class BGGDualModule(CombinatorialFreeModule):
    r"""
    The dual module `M^{\vee}` in the BGG Category `\mathcal{O}`.

    Let `\tau` be the transpose map of a semisimple (finite dimensional)
    Lie algebra `\mathfrak{g}` over a field `R`. Let `M \in \mathcal{O}`.
    The *BGG dual module* is the `R`-module `M^{\vee} :=
    \bigoplus_{\lambda} M_{\lambda}^*` which has a `U(\mathfrak{g})`-module
    structure given by

    .. MATH::

        x \cdot \phi(v) := \phi(\tau(x) \cdot v),

    which is also a weight module with the same grading as `M`.

    The basis we chose to work with here is the natural dual basis to the
    distinguished basis `B` of `M`. That is, we define the dual function
    to `b` as `\phi_b(c) = \delta_{bc}`.

    EXAMPLES::

        sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
        sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
        sage: M = g.verma_module(2*La[1])
        sage: Mc = M.dual()
        sage: B = Mc.basis()
        sage: it = iter(B)
        sage: elts = [next(it) for _ in range(7)]; elts
        [v[2*Lambda[1]]^*,
         f[-alpha[1]]*v[2*Lambda[1]]^*,
         f[-alpha[1]]^2*v[2*Lambda[1]]^*,
         f[-alpha[1]]^3*v[2*Lambda[1]]^*,
         f[-alpha[1]]^4*v[2*Lambda[1]]^*,
         f[-alpha[1]]^5*v[2*Lambda[1]]^*,
         f[-alpha[1]]^6*v[2*Lambda[1]]^*]
        sage: e, h, f = g.pbw_basis().algebra_generators()
        sage: [f * vec for vec in elts]
        [2*f[-alpha[1]]*v[2*Lambda[1]]^*,
         2*f[-alpha[1]]^2*v[2*Lambda[1]]^*,
         0,
         -4*f[-alpha[1]]^4*v[2*Lambda[1]]^*,
         -10*f[-alpha[1]]^5*v[2*Lambda[1]]^*,
         -18*f[-alpha[1]]^6*v[2*Lambda[1]]^*,
         -28*f[-alpha[1]]^7*v[2*Lambda[1]]^*]
        sage: [e * vec for vec in elts]
        [0,
         v[2*Lambda[1]]^*,
         f[-alpha[1]]*v[2*Lambda[1]]^*,
         f[-alpha[1]]^2*v[2*Lambda[1]]^*,
         f[-alpha[1]]^3*v[2*Lambda[1]]^*,
         f[-alpha[1]]^4*v[2*Lambda[1]]^*,
         f[-alpha[1]]^5*v[2*Lambda[1]]^*]
        sage: [h * vec for vec in elts]
        [2*v[2*Lambda[1]]^*,
         0,
         -2*f[-alpha[1]]^2*v[2*Lambda[1]]^*,
         -4*f[-alpha[1]]^3*v[2*Lambda[1]]^*,
         -6*f[-alpha[1]]^4*v[2*Lambda[1]]^*,
         -8*f[-alpha[1]]^5*v[2*Lambda[1]]^*,
         -10*f[-alpha[1]]^6*v[2*Lambda[1]]^*]

    REFERENCES:

    - [Humphreys08]_
    """
    def __init__(self, module) -> None:
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['B', 2])
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: M = g.verma_module(2*La[1] + La[2])
            sage: Mc = M.dual()
            sage: TestSuite(Mc).run()

            sage: M = g.verma_module(2/3*La[1] - 3/5*La[2])
            sage: Mc = M.dual()
            sage: TestSuite(Mc).run()
        """
        self._module = module
        self._g = module.lie_algebra()
        self._pbw = self._g.pbw_basis()
        base_ring = module.base_ring()
        indices = module.indices()
        category = module.category()
        CombinatorialFreeModule.__init__(self, base_ring, indices, category=category,
                                         **module.print_options())

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: M = g.verma_module(2*La[1])
            sage: M.dual()
            BGG Dual of Verma module with highest weight 2*Lambda[1] of
             Lie algebra of ['A', 1] in the Chevalley basis
        """
        return "BGG Dual of " + repr(self._module)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: M = g.verma_module(2*La[1])
            sage: Mc = M.dual()
            sage: latex(Mc)
            { M_{2 \Lambda_{1}} }^{\vee}
        """
        from sage.misc.latex import latex
        return "{" + latex(self._module) + "}^{\\vee}"

    def _repr_generator(self, m):
        r"""
        Return a string representation of the generator indexed by ``m``.

        EXAMPLES::

            sage: g = lie_algebras.sp(QQ, 4)
            sage: La = g.cartan_type().root_system().ambient_space().fundamental_weights()
            sage: Mc = g.verma_module(La[1] + 3/7*La[2]).dual()
            sage: f1, f2 = g.f()
            sage: x = g.pbw_basis()(g([f1, [f1, f2]]))
            sage: v = x * Mc.highest_weight_vector()
            sage: Mc._repr_generator(v.leading_support())
            'f[-alpha[1]]*f[-alpha[1] - alpha[2]]*v[(10/7, 3/7)]^*'
        """
        return self._module._repr_generator(m) + "^*"

    def _latex_generator(self, m):
        r"""
        Return a latex representation of the generator indexed by ``m``.

        EXAMPLES::

            sage: g = lie_algebras.sp(QQ, 4)
            sage: La = g.cartan_type().root_system().ambient_space().fundamental_weights()
            sage: Mc = g.verma_module(La[1] + 3/7*La[2]).dual()
            sage: f1, f2 = g.f()
            sage: x = g.pbw_basis()(g([f1, [f1, f2]]))
            sage: v = x * Mc.highest_weight_vector()
            sage: Mc._latex_generator(v.leading_support())
            { f_{-\alpha_{1}} f_{-\alpha_{1} - \alpha_{2}} v_{\frac{10}{7} e_{0} + \frac{3}{7} e_{1}} }^{\vee}
        """
        return "{" + self._module._latex_generator(m) + "}^{\\vee}"

    _repr_term = _repr_generator
    _latex_term = _latex_generator

    def degree_on_basis(self, m):
        r"""
        Return the degree of the basis element indexed by ``m``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['D', 5])
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: M = g.verma_module(La[1] + La[4] - 1/3*La[5])
            sage: Mc = M.dual()
            sage: elt = Mc.an_element(); elt
            f[-alpha[5]]^2*f[-alpha[4]]^2*f[-alpha[3]]^3*v[Lambda[1] + Lambda[4] - 1/3*Lambda[5]]^*
             + 2*f[-alpha[5]]*v[Lambda[1] + Lambda[4] - 1/3*Lambda[5]]^*
             + 3*f[-alpha[4]]*v[Lambda[1] + Lambda[4] - 1/3*Lambda[5]]^*
             + v[Lambda[1] + Lambda[4] - 1/3*Lambda[5]]^*
            sage: [M.degree_on_basis(m) for m in elt.support()]
            [Lambda[1] + 3*Lambda[2] - 2*Lambda[3] - 4/3*Lambda[5],
             Lambda[1] + Lambda[4] - 1/3*Lambda[5],
             Lambda[1] + Lambda[3] + Lambda[4] - 7/3*Lambda[5],
             Lambda[1] + Lambda[3] - Lambda[4] - 1/3*Lambda[5]]
        """
        return self._module.degree_on_basis(m)

    def highest_weight(self):
        r"""
        Return the highest weight of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 7])
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: M = g.verma_module(2*La[1] + 5/3*La[4] - 3*La[6])
            sage: Mc = M.dual()
            sage: Mc.highest_weight()
            2*Lambda[1] + 5/3*Lambda[4] - 3*Lambda[6]
        """
        return self._module.highest_weight()

    def highest_weight_vector(self):
        r"""
        Return the highest weight vector of ``self`` (assuming the
        defining module defines such a vector).

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: M = g.verma_module(2*La[1])
            sage: Mc = M.dual()
            sage: Mc.highest_weight_vector()
            v[2*Lambda[1]]^*
        """
        hwv = self._module.highest_weight_vector()
        return self.element_class(self, hwv.monomial_coefficients(copy=False))

    def lie_algebra(self):
        r"""
        Return the underlying Lie algebra of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['B', 3])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: M = g.verma_module(2*La[1] + La[3])
            sage: Mc = M.dual()
            sage: Mc.lie_algebra() is g
            True
        """
        return self._g

    def dual(self):
        r"""
        Return the dual module of ``self``.

        In Category `\mathcal{O}`, we have `(M^{\vee})^{\vee} \cong M`, so
        we return the defining module `M` of `M^{\vee}`.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['F', 4])
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: M = g.verma_module(La[1] - 5/3*La[2] + 3*La[4])
            sage: Mc = M.dual()
            sage: Mc.dual() is M
            True
        """
        return self._module

    @cached_method
    def _lie_algebra_on_basis(self, b, m):
        r"""
        Return the action of the Lie algebra basis element indexed by ``b``
        on the basis element of ``self`` indexed by ``m``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['B', 2])
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: M = g.verma_module(La[1])
            sage: Mc = M.dual()
            sage: it = iter(Mc.basis())
            sage: list(g.basis())
            [E[alpha[2]], E[alpha[1]], E[alpha[1] + alpha[2]], E[alpha[1] + 2*alpha[2]],
             h1, h2,
             E[-alpha[2]], E[-alpha[1]], E[-alpha[1] - alpha[2]], E[-alpha[1] - 2*alpha[2]]]
            sage: for _ in range(3):
            ....:     m = next(it).leading_support()
            ....:     print(m, [Mc._lie_algebra_on_basis(k, m) for k in g.basis().keys()])
            1 [0, 0, 0, 0, v[Lambda[1]]^*, 0, 0, f[-alpha[1]]*v[Lambda[1]]^*,
               -2*f[-alpha[2]]*f[-alpha[1]]*v[Lambda[1]]^* + 2*f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*,
               -2*f[-alpha[2]]^2*f[-alpha[1]]*v[Lambda[1]]^*
                + 2*f[-alpha[2]]*f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*
                + f[-alpha[1] - 2*alpha[2]]*v[Lambda[1]]^*]
            f[-alpha[2]] [v[Lambda[1]]^*, 0, 0, 0, 2*f[-alpha[2]]*v[Lambda[1]]^*,
                -2*f[-alpha[2]]*v[Lambda[1]]^*, -2*f[-alpha[2]]^2*v[Lambda[1]]^*,
                f[-alpha[2]]*f[-alpha[1]]*v[Lambda[1]]^* + f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*,
                -4*f[-alpha[2]]^2*f[-alpha[1]]*v[Lambda[1]]^* - f[-alpha[1] - 2*alpha[2]]*v[Lambda[1]]^*,
                -6*f[-alpha[2]]^3*f[-alpha[1]]*v[Lambda[1]]^* + 2*f[-alpha[2]]^2*f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*]
            f[-alpha[1]] [0, v[Lambda[1]]^*, 0, 0, -f[-alpha[1]]*v[Lambda[1]]^*,
                2*f[-alpha[1]]*v[Lambda[1]]^*,
                2*f[-alpha[2]]*f[-alpha[1]]*v[Lambda[1]]^* - 2*f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*,
                0, 0, f[-alpha[1]]*f[-alpha[1] - 2*alpha[2]]*v[Lambda[1]]^* + 2*f[-alpha[1] - alpha[2]]^2*v[Lambda[1]]^*]
        """
        al = self._g.degree_on_basis(b)
        wt = self.degree_on_basis(m)
        if al == 0:  # b is indexing part of the Cartan subalgebra
            # We are assuming b is part of the coroot lattice.
            # FIXME: Add something at the category level to return this.
            ac = b
            return self.term(m, wt.scalar(ac))

        # TODO: Avoid calling homogeneous_component_basis() as the result is not cached
        gens = self._module.homogeneous_component_basis(wt + al)
        elt = self._g.basis()[b]
        # TODO: Determine if we can meaningfully store these results.
        # Computing gens is ~1/3 of the computation and vecs is ~2/3.
        vecs = {g.leading_support(): elt.transpose() * g for g in gens}
        return self.element_class(self, {k: c for k, v in vecs.items() if (c := v[m])})

    def _pbw_monomial_on_basis(self, p, m):
        r"""
        Return the action of the PBW monomial indexed by ``p`` on the basis
        element of ``self`` indexed by ``m``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: PBW = g.pbw_basis()
            sage: e, h, f = PBW.algebra_generators()
            sage: M = g.verma_module(2*La[1])
            sage: Mc = M.dual()
            sage: v = Mc.highest_weight_vector()
            sage: Mc._pbw_monomial_on_basis((e*f^2).leading_support(), v.leading_support())
            4*f[-alpha[1]]*v[2*Lambda[1]]^*
            sage: B = Mc.basis()
            sage: it = iter(B)
            sage: elts = [next(it) for _ in range(7)]; elts
            [v[2*Lambda[1]]^*,
             f[-alpha[1]]*v[2*Lambda[1]]^*,
             f[-alpha[1]]^2*v[2*Lambda[1]]^*,
             f[-alpha[1]]^3*v[2*Lambda[1]]^*,
             f[-alpha[1]]^4*v[2*Lambda[1]]^*,
             f[-alpha[1]]^5*v[2*Lambda[1]]^*,
             f[-alpha[1]]^6*v[2*Lambda[1]]^*]
        """
        ret = self.monomial(m)
        for b, exp in reversed(p._sorted_items()):
            for _ in range(exp):
                ret = self.linear_combination((self._lie_algebra_on_basis(b, m), mc)
                                              for m, mc in ret._monomial_coefficients.items())
        return ret

    class Element(CombinatorialFreeModule.Element):
        def _acted_upon_(self, scalar, self_on_left=False):
            r"""
            Return the action of ``scalar`` on ``self``.

            EXAMPLES::

                sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
                sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
                sage: PBW = g.pbw_basis()
                sage: e, h, f = PBW.algebra_generators()
                sage: M = g.verma_module(2*La[1])
                sage: Mc = M.dual()
                sage: v = Mc.highest_weight_vector()
                sage: (h*e^2*f^2) * v
                8*v[2*Lambda[1]]^*
                sage: g.casimir_element(UEA=PBW) * v
                v[2*Lambda[1]]^*
                sage: 5 * v
                5*v[2*Lambda[1]]^*
            """
            P = self.parent()
            # Check for scalars first
            if scalar in P.base_ring():
                # Don't have this be a super call
                return CombinatorialFreeModule.Element._acted_upon_(self, scalar, self_on_left)

            # Check for Lie algebra elements
            try:
                scalar = P._g(scalar)
            except (ValueError, TypeError):
                pass
            if scalar.parent() is P._g:
                if self_on_left:  # only implemented as a left module
                    return None
                mc = scalar.monomial_coefficients(copy=False)
                return P.linear_combination((P._lie_algebra_on_basis(b, m), bc * mc)
                                            for b, bc in mc.items()
                                            for m, mc in self._monomial_coefficients.items())

            # Check for PBW elements
            try:
                scalar = P._pbw(scalar)
            except (ValueError, TypeError):
                # Cannot be made into a PBW element, so propagate it up
                return CombinatorialFreeModule.Element._acted_upon_(self,
                        scalar, self_on_left)

            # We only implement x * self, i.e., as a left module
            if self_on_left:
                return None

            mc = scalar.monomial_coefficients(copy=False)
            return P.linear_combination((P._pbw_monomial_on_basis(p, m), pc * mc)
                                        for p, pc in mc.items()
                                        for m, mc in self._monomial_coefficients.items())


#####################################################################
# Simple modules


# This is an abuse as the monoid is not free.
# TODO: Rewrite this (or the indexed monoid class) to use explicit vectors
#    since we only want to consider ordered elements.
# Note, such a rewrite would force the Lie algebra to be finite dimensional.
class SimpleModuleIndices(IndexedFreeAbelianMonoid):
    r"""
    The indices of the basis for a simple `U(\mathfrak{g})`-module.

    .. NOTE::

        The current implementation assumes the Lie algebra `\mathfrak{g}`
        is finite dimensional.
    """
    # This is only necessary because of the IndexedMonoid.__classcall__.
    @staticmethod
    def __classcall__(cls, simple, prefix='f', **kwds):
        r"""
        Normalize input to ensure a unique representation.

        TESTS::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1] + La[3])
            sage: from sage.algebras.lie_algebras.bgg_dual_module import SimpleModuleIndices
            sage: SimpleModuleIndices(L) is L._indices
            True
        """
        return super(IndexedMonoid, cls).__classcall__(cls, simple, prefix=prefix, **kwds)

    def __init__(self, simple, prefix, category=None, **kwds) -> None:
        r"""
        Initialize ``self``.

        TESTS::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: I = g.simple_module(2*La[1] + La[2]).indices()
            sage: TestSuite(I).run()

            sage: I = g.simple_module(2*La[1] - 1/3*La[2]).indices()
            sage: TestSuite(I).run(max_runs=150)  # long time
        """
        self._simple = simple
        self._g = simple.lie_algebra()
        self._reached_max_depth = False
        # Below was mostly copied from IndexedMonoid.__init__()
        self._indices = FiniteEnumeratedSet(self._g._negative_half_index_set())
        category = Monoids().or_subcategory(category)
        category = category & EnumeratedSets()
        category = category.FinitelyGeneratedAsMagma()
        if self._simple._dom_int:
            category = category.Finite()
        else:
            category = category.Infinite()
        Parent.__init__(self, category=category)

        # ignore the optional 'key' since it only affects CachedRepresentation
        kwds.pop('key', None)
        sorting_key = kwds.pop('sorting_key', self._simple._pbw._monoid_key)
        IndexedGenerators.__init__(self, self._indices, prefix, sorting_key=sorting_key, **kwds)

        self._sorted_supp = sorted(self._g._negative_half_index_set(), key=self._simple._pbw._basis_key,
                                   reverse=self.print_options()['sorting_reverse'])
        self._basis = {self.one(): self._simple._ambient.highest_weight_vector()}
        self._lead_supp_to_index = {self._simple._ambient.highest_weight_vector().leading_support(): self.one()}
        # This is used for iteration and keeps track of the current depth
        self._basis_by_depth = [dict(self._basis)]
        # The basis is given as a list of indices corresponding to basis vectors in self._basis
        self._weight_space_bases = {self._simple.highest_weight(): [self.one()]}

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        The only element we can quickly guarantee is in ``self`` is 1,
        so we return this.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: I = g.simple_module(2*La[1] + La[2]).indices()
            sage: I._an_element_()
            1
        """
        return self.one()

    def _weight_max_depth(self, mu):
        r"""
        Return the maximum depth of the weight ``mu``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: P = g.cartan_type().root_system().weight_lattice()
            sage: La = P.fundamental_weights()
            sage: al = P.simple_roots()
            sage: wt = 2*La[1] + La[2]
            sage: I = g.simple_module(wt).indices()
            sage: I._weight_max_depth(wt)
            0
            sage: I._weight_max_depth(wt + al[2]) is None
            True
            sage: I._weight_max_depth(wt - 2*al[2] - 5*al[4] - 3*al[6])
            10

            sage: g = LieAlgebra(QQ, cartan_type=['F', 4])
            sage: P = g.cartan_type().root_system().weight_space()
            sage: La = P.fundamental_weights()
            sage: al = P.simple_roots()
            sage: wt = 2*La[1] - 3/2*La[2]
            sage: I = g.simple_module(wt).indices()
            sage: I._weight_max_depth(wt)
            0
            sage: I._weight_max_depth(wt + al[2]) is None
            True
            sage: I._weight_max_depth(wt - 2*al[2] - 3*al[4])
            5
            sage: I._weight_max_depth(wt - 2/3*al[1]) is None
            True
        """
        al = (self._simple.highest_weight() - mu)._to_root_vector()
        if any(c not in ZZ or c < 0 for c in al):
            return None
        return sum(al)

    def weight_space_basis(self, mu):
        r"""
        Return the indices of the ``mu`` weight space basis elements.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: P = g.cartan_type().root_system().weight_lattice()
            sage: La = P.fundamental_weights()
            sage: al = P.simple_roots()
            sage: wt = -3*La[1] + 3*La[2]
            sage: I = g.simple_module(wt).indices()
            sage: I.weight_space_basis(wt)
            [1]
            sage: I.weight_space_basis(wt - al[1])
            [f[-alpha[1]]]
            sage: I.weight_space_basis(wt - al[2])
            [f[-alpha[2]]]
            sage: I.weight_space_basis(wt - al[1] - al[2])
            [f[-alpha[1] - alpha[2]], f[-alpha[2]]*f[-alpha[1]]]
            sage: I.weight_space_basis(wt - 4*al[1])
            [f[-alpha[1]]^4]
            sage: I.weight_space_basis(wt - 4*al[2])
            []
        """
        if self._reached_max_depth:
            return self._weight_space_bases.get(mu, [])

        max_depth = self._weight_max_depth(mu)
        while max_depth >= len(self._basis_by_depth):
            if self._reached_max_depth:  # we've already reached everything
                break
            self._construct_next_level()
        return self._weight_space_bases.get(mu, [])

    def __contains__(self, m) -> bool:
        r"""
        Check if ``m`` is contained in ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['G', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1])
            sage: I = L.indices()
            sage: I.one() in I
            True
            sage: it = iter(I)
            sage: for _ in range(3):
            ....:     elt = next(it)
            ....:     print(elt, elt in I)
            1 True
            f[-alpha[1]] True
            f[-alpha[1] - alpha[2]] True
            sage: gens = list(I.gens()); gens
            [f[-alpha[2]],
             f[-alpha[1]],
             f[-alpha[1] - alpha[2]],
             f[-2*alpha[1] - alpha[2]],
             f[-3*alpha[1] - alpha[2]],
             f[-3*alpha[1] - 2*alpha[2]]]
            sage: gens[1] in I
            True
            sage: gens[0] * gens[1] in I
            False
            sage: gens[2] in I
            True
            sage: gens[0]^10 in I
            False
            sage: gens[5]^6 * gens[2]^10 in I
            False
        """
        if not isinstance(m, self.Element) or m.parent() is not self:
            return False
        depth = m.length()
        while depth >= len(self._basis_by_depth):
            if self._reached_max_depth:  # we've already reached everything
                break
            self._construct_next_level()
        return m in self._basis

    def __iter__(self):
        r"""
        Iterate over ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['B', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[2])
            sage: I = L.indices()
            sage: list(I)
            [1, f[-alpha[2]], f[-alpha[1] - alpha[2]], f[-alpha[1] - 2*alpha[2]]]

            sage: L = g.simple_module(La[1]-La[2])
            sage: I = L.indices()
            sage: it = iter(I)
            sage: [next(it) for _ in range(6)]
            [1, f[-alpha[2]], f[-alpha[1]], f[-alpha[1] - alpha[2]],
             f[-alpha[1] - 2*alpha[2]], f[-alpha[2]]^2]
        """
        depth = 0
        while True:
            while depth >= len(self._basis_by_depth):
                if self._reached_max_depth:  # we've already reached everything
                    return
                self._construct_next_level()
            yield from self._basis_by_depth[depth]
            depth += 1

    def _construct_next_level(self):
        r"""
        Construct the image for the next level of ``self``.

        ALGORITHM:

        For each image vector of `f_{\beta_1}^{b_1} \cdots f_{\beta_k}^{b_k}
        v_{\lambda}` at the current depth `b_1 + \cdots b_k`, consider the
        image under multiplication by every generator `f_{\alpha}` for
        all `\alpha \leq \beta_1` in the fixed PBW ordering of the (dual)
        Verma module.

        .. TODO::

            Avoid unnecessary computations by using the corresponding
            (combinatorial) crystal.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['C', 3])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1])
            sage: I = L.indices()
            sage: len(I._basis)
            1
            sage: I._construct_next_level()
            sage: len(I._basis)
            6
            sage: I._reached_max_depth
            False
            sage: I._construct_next_level()  # long time
            sage: len(I._basis)  # long time
            6
            sage: I._reached_max_depth  # long time
            True
            sage: I._construct_next_level()  # long time
        """
        if self._reached_max_depth:  # we've already reached everything
            return  # so nothing more to do

        gens = self._g.basis()
        next_level = {}
        R = self._g.base_ring()
        ambient = self._simple._ambient
        pbw = ambient._pbw
        for m, vec in self._basis_by_depth[-1].items():
            # find the first support index
            ind = len(self._sorted_supp)
            for i, ls in enumerate(self._sorted_supp):
                if ls in m._monomial:
                    ind = i + 1
                    break

            for ls in self._sorted_supp[:ind]:
                mp = dict(m._monomial)  # make a (shallow) copy
                mp[ls] = mp.get(ls, 0) + 1
                key = self.element_class(self, mp)
                new_vec = gens[ls] * vec
                if not new_vec:
                    continue
                # Echelonize the corresponding weight space
                mu = ambient.degree_on_basis(key)
                if mu not in self._weight_space_bases:
                    # the only vector in the weight space
                    self._weight_space_bases[mu] = [key]
                    next_level[key] = new_vec
                    self._basis[key] = next_level[key]
                    lead_supp = next_level[key].trailing_support(key=pbw._monomial_key)
                    self._lead_supp_to_index[lead_supp] = key
                    continue

                supp = set()
                wt_basis = [self._basis[k] for k in self._weight_space_bases[mu]]
                wt_basis.append(new_vec)
                for b in wt_basis:
                    supp.update(b.support())
                supp = sorted(supp, key=pbw._monomial_key)
                mat = matrix(R, [[b[s] for s in supp] for b in wt_basis])
                mat.echelonize()
                for i, k in enumerate(self._weight_space_bases[mu]):
                    data = {supp[ind]: R(c) for ind, c in mat[i].items() if c}
                    self._basis[k] = ambient.element_class(ambient, data)
                i = mat.nrows() - 1
                data = {supp[ind]: R(c) for ind, c in mat[i].items() if c}
                if data:
                    next_level[key] = ambient.element_class(ambient, data)
                    self._basis[key] = next_level[key]
                    lead_supp = next_level[key].trailing_support(key=pbw._monomial_key)
                    self._lead_supp_to_index[lead_supp] = key
                    self._weight_space_bases[mu].append(key)

        if not next_level:
            self._reached_max_depth = True
            return
        self._basis_by_depth.append(next_level)

    @cached_method
    def cardinality(self):
        r"""
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1]+La[4])
            sage: L._indices.cardinality()
            51975
        """
        if self._simple._dom_int:
            weight = self._simple.highest_weight()
            Phi = self._g.cartan_type().root_system()
            P = Phi.weight_lattice()
            coroots = Phi.root_lattice().simple_coroots()
            la = P._from_dict({i: weight.scalar(ac) for i, ac in coroots.items()})
            from sage.combinat.crystals.monomial_crystals import (
                CrystalOfNakajimaMonomials,
            )
            return CrystalOfNakajimaMonomials(la).cardinality()
        from sage.rings.infinity import infinity
        return infinity


class SimpleModule(ModulePrinting, CombinatorialFreeModule):
    r"""
    Return the simple module `L_{\lambda}` as the image of the natural
    morphism `\phi: M_{\lambda} \to M_{\lambda}^{\vee}`.
    """
    @staticmethod
    def __classcall_private__(cls, g, weight, *args, **kwds):
        r"""
        Normalize input to ensure a unique representation and return
        the correct type.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['E', 6])
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: type(g.simple_module(La[1] + La[2]))
            <class 'sage.algebras.lie_algebras.bgg_dual_module.FiniteDimensionalSimpleModule_with_category'>
            sage: type(g.simple_module(La[1] - La[2]))
            <class 'sage.algebras.lie_algebras.bgg_dual_module.SimpleModule_with_category'>
            sage: type(g.simple_module(La[1] + 3/2*La[2]))
            <class 'sage.algebras.lie_algebras.bgg_dual_module.SimpleModule_with_category'>
        """
        if weight.is_dominant_weight():
            return FiniteDimensionalSimpleModule(g, weight, *args, **kwds)
        return super().__classcall__(cls, g, weight, *args, **kwds)

    def __init__(self, g, weight, prefix='f', basis_key=None, **kwds) -> None:
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['G', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1] + La[2])
            sage: TestSuite(L).run()

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1] - La[2])
            sage: TestSuite(L).run()
        """
        self._g = g
        self._weight = weight
        self._dom_int = weight.is_dominant_weight()
        self._verma = g.verma_module(weight, basis_key=basis_key)
        self._ambient = self._verma.dual()
        self._pbw = self._verma.pbw_basis()
        base_ring = self._g.base_ring()
        indices = SimpleModuleIndices(self, prefix=prefix, **kwds)
        category = self._ambient.category().Subobjects()
        if self._dom_int:
            category = category.FiniteDimensional()
        ModulePrinting.__init__(self, 'u')
        CombinatorialFreeModule.__init__(self, base_ring, indices, category=category,
                                         **self._ambient.print_options())

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: g.simple_module(2*La[1])
            Simple module with highest weight 2*Lambda[1] of
             Lie algebra of ['A', 1] in the Chevalley basis
        """
        return "Simple module with highest weight {} of {}".format(self._weight, self._g)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(2*La[1])
            sage: latex(L)
            L_{2 \Lambda_{1}}
        """
        from sage.misc.latex import latex
        return "L_{{{}}}".format(latex(self._weight))

    def ambient(self):
        r"""
        Return the ambient module of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['G', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(2*La[1])
            sage: L.ambient()
            BGG Dual of Verma module with highest weight 2*Lambda[1] of
             Lie algebra of ['G', 2] in the Chevalley basis
        """
        return self._ambient

    @lazy_attribute
    def lift(self):
        r"""
        Return the lift map of ``self`` to the ambient dual Verma module.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['G', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1])
            sage: [L.lift(b) for b in L.basis()]  # long time
            [v[Lambda[1]]^*,
             f[-alpha[1]]*v[Lambda[1]]^*,
             f[-alpha[2]]*f[-alpha[1]]*v[Lambda[1]]^* - f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*,
             f[-alpha[1]]*f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*
              + f[-2*alpha[1] - alpha[2]]*v[Lambda[1]]^*,
             f[-alpha[1]]^2*f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*
              + f[-alpha[1]]*f[-2*alpha[1] - alpha[2]]*v[Lambda[1]]^*
              + 1/2*f[-3*alpha[1] - alpha[2]]*v[Lambda[1]]^*,
             f[-alpha[2]]*f[-alpha[1]]^2*f[-alpha[1] - alpha[2]]*v[Lambda[1]]^*
              + f[-alpha[2]]*f[-alpha[1]]*f[-2*alpha[1] - alpha[2]]*v[Lambda[1]]^*
              + 1/2*f[-alpha[2]]*f[-3*alpha[1] - alpha[2]]*v[Lambda[1]]^*
              - f[-alpha[1] - alpha[2]]*f[-2*alpha[1] - alpha[2]]*v[Lambda[1]]^*
              + 1/2*f[-3*alpha[1] - 2*alpha[2]]*v[Lambda[1]]^*,
             f[-alpha[1]]*f[-alpha[1] - alpha[2]]*f[-2*alpha[1] - alpha[2]]*v[Lambda[1]]^*
              - 1/2*f[-alpha[1]]*f[-3*alpha[1] - 2*alpha[2]]*v[Lambda[1]]^*
              - 1/2*f[-alpha[1] - alpha[2]]*f[-3*alpha[1] - alpha[2]]*v[Lambda[1]]^*
              + f[-2*alpha[1] - alpha[2]]^2*v[Lambda[1]]^*]
        """
        return self.module_morphism(self._lift_on_basis, codomain=self._ambient, unitriangular="upper")

    def retract(self, x):
        r"""
        Return the retraction of ``x`` in ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(2*La[1])
            sage: L.retract(L.lift(sum(L.basis())))
            f[-alpha[1]]^2*u[2*Lambda[1]] + f[-alpha[1]]*f[-alpha[1] - alpha[2]]*u[2*Lambda[1]]
             + f[-alpha[1] - alpha[2]]^2*u[2*Lambda[1]] + f[-alpha[1]]*u[2*Lambda[1]]
             + f[-alpha[1] - alpha[2]]*u[2*Lambda[1]] + u[2*Lambda[1]]
            sage: B = list(L.basis())
            sage: L.retract(3/2*L.lift(B[0]) - L.lift(B[2]) - 10/3*L.lift(B[3]))
            -10/3*f[-alpha[1]]^2*u[2*Lambda[1]]
             - f[-alpha[1] - alpha[2]]*u[2*Lambda[1]]
             + 3/2*u[2*Lambda[1]]
        """
        supp = sorted(x.support(), key=self._pbw._monomial_key)
        data = x.monomial_coefficients(copy=True)  # this is destructive to data
        R = self.base_ring()
        ret = {}
        for ls in supp:
            if ls not in data:
                continue
            if ls not in self._indices._lead_supp_to_index:
                mu = self._ambient.degree_on_basis(ls)
                # this will guarantee the computation is correct
                self._indices.weight_space_basis(mu)
            if ls not in self._indices._lead_supp_to_index:
                raise ValueError(f"not an element of the simple module of weight {self._weight}")
            key = self._indices._lead_supp_to_index[ls]
            vec = self._indices._basis[key]
            coeff = R(data[ls] / vec[ls])
            iaxpy(-coeff, vec._monomial_coefficients, data)
            ret[key] = coeff
        return self.element_class(self, ret)

    def _lift_on_basis(self, m):
        r"""
        Return the lift of the basis element indexed by ``m``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 1])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(2*La[1])
            sage: I = L.indices()
            sage: gen = list(I.gens())[0]
            sage: L._lift_on_basis(gen^2)
            4*f[-alpha[1]]^2*v[2*Lambda[1]]^*
            sage: L._lift_on_basis(gen^3)
            Traceback (most recent call last):
            ...
            ValueError: f[-alpha[1]]^3 does not index a basis element
        """
        # This builds the result up to the necessary depth
        if m not in self._indices:
            raise ValueError(f"{m} does not index a basis element")
        return self._indices._basis[m]

    def dual(self) -> Self:
        r"""
        Return the dual module of ``self``, which is ``self`` since simple
        modules are self-dual.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['B', 4])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(2*La[1] + 3*La[4])
            sage: L.dual() is L
            True
        """
        return self

    def highest_weight(self):
        r"""
        Return the highest weight of ``self``.

        EXAMPLES::

            sage: g = lie_algebras.so(QQ, 7)
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1] + La[2])
            sage: L.highest_weight()
            Lambda[1] + Lambda[2]
        """
        return self._weight

    @cached_method
    def highest_weight_vector(self):
        r"""
        Return the highest weight vector of ``self``.

        EXAMPLES::

            sage: g = lie_algebras.sp(QQ, 6)
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1] + La[2])
            sage: L.highest_weight_vector()
            u[Lambda[1] + Lambda[2]]
        """
        one = self.base_ring().one()
        return self._from_dict({self._indices.one(): one},
                               remove_zeros=False, coerce=False)

    def lie_algebra(self):
        r"""
        Return the underlying Lie algebra of ``self``.

        EXAMPLES::

            sage: g = lie_algebras.so(QQ, 9)
            sage: La = g.cartan_type().root_system().weight_space().fundamental_weights()
            sage: L = g.simple_module(La[3] - 1/2*La[1])
            sage: L.lie_algebra()
            Lie algebra of ['B', 4] in the Chevalley basis
        """
        return self._g

    def pbw_basis(self):
        r"""
        Return the PBW basis of the underlying Lie algebra
        used to define ``self``.

        EXAMPLES::

            sage: g = lie_algebras.so(QQ, 8)
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[2] - 2*La[3])
            sage: L.pbw_basis()
            Universal enveloping algebra of Lie algebra of ['D', 4] in the Chevalley basis
             in the Poincare-Birkhoff-Witt basis
        """
        return self._pbw

    def homogeneous_component_basis(self, mu):
        r"""
        Return a basis for the ``mu`` weight space of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: P = g.cartan_type().root_system().weight_lattice()
            sage: La = P.fundamental_weights()
            sage: la = La[1] + La[2]
            sage: L = g.simple_module(la)
            sage: from itertools import product
            sage: al = P.simple_roots()
            sage: for wts in product(range(4), repeat=2):
            ....:     mu = la - wts[0] * al[1] - wts[1] * al[2]
            ....:     print(mu)
            ....:     print(L.homogeneous_component_basis(mu))
            Lambda[1] + Lambda[2]
            Family (u[Lambda[1] + Lambda[2]],)
            2*Lambda[1] - Lambda[2]
            Family (f[-alpha[2]]*u[Lambda[1] + Lambda[2]],)
            3*Lambda[1] - 3*Lambda[2]
            Family ()
            4*Lambda[1] - 5*Lambda[2]
            Family ()
            -Lambda[1] + 2*Lambda[2]
            Family (f[-alpha[1]]*u[Lambda[1] + Lambda[2]],)
            0
            Family (f[-alpha[1] - alpha[2]]*u[Lambda[1] + Lambda[2]], f[-alpha[2]]*f[-alpha[1]]*u[Lambda[1] + Lambda[2]])
            Lambda[1] - 2*Lambda[2]
            Family (f[-alpha[2]]*f[-alpha[1] - alpha[2]]*u[Lambda[1] + Lambda[2]],)
            2*Lambda[1] - 4*Lambda[2]
            Family ()
            -3*Lambda[1] + 3*Lambda[2]
            Family ()
            -2*Lambda[1] + Lambda[2]
            Family (f[-alpha[1]]*f[-alpha[1] - alpha[2]]*u[Lambda[1] + Lambda[2]],)
            -Lambda[1] - Lambda[2]
            Family (f[-alpha[1] - alpha[2]]^2*u[Lambda[1] + Lambda[2]],)
            -3*Lambda[2]
            Family ()
            -5*Lambda[1] + 4*Lambda[2]
            Family ()
            -4*Lambda[1] + 2*Lambda[2]
            Family ()
            -3*Lambda[1]
            Family ()
            -2*Lambda[1] - 2*Lambda[2]
            Family ()
        """
        return Family([self.monomial(b) for b in self._indices.weight_space_basis(mu)])

    weight_space_basis = homogeneous_component_basis

    class Element(CombinatorialFreeModule.Element):
        def _acted_upon_(self, scalar, self_on_left=True):
            r"""
            Return the action of ``scalar`` on ``self``.

            EXAMPLES::

                sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
                sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
                sage: L = g.simple_module(La[1] + La[2])
                sage: v = L.highest_weight_vector(); v
                u[Lambda[1] + Lambda[2]]
                sage: f1, f2 = g.pbw_basis().f()
                sage: 5 * v
                5*u[Lambda[1] + Lambda[2]]
                sage: f1 * f2 * v
                f[-alpha[2]]*f[-alpha[1]]*u[Lambda[1] + Lambda[2]]
                 + f[-alpha[1] - alpha[2]]*u[Lambda[1] + Lambda[2]]
                sage: f2 * f1 * v
                -f[-alpha[2]]*f[-alpha[1]]*u[Lambda[1] + Lambda[2]]
                 + 2*f[-alpha[1] - alpha[2]]*u[Lambda[1] + Lambda[2]]
                sage: f2 * f2 * f1 * v
                -2*f[-alpha[2]]*f[-alpha[1] - alpha[2]]*u[Lambda[1] + Lambda[2]]
                sage: f1 * f2 * f1 * v
                f[-alpha[1]]*f[-alpha[1] - alpha[2]]*u[Lambda[1] + Lambda[2]]
                sage: f2 * f1 * f2 * f1 * v
                f[-alpha[1] - alpha[2]]^2*u[Lambda[1] + Lambda[2]]
                sage: f1 * f2 * f2 * f1 * v
                2*f[-alpha[1] - alpha[2]]^2*u[Lambda[1] + Lambda[2]]
            """
            # check for scalars first
            ret = CombinatorialFreeModule.Element._acted_upon_(self, scalar, self_on_left)
            if ret is not None:
                return ret

            if self_on_left:  # this is a left module action
                return None

            P = self.parent()
            return P.retract(scalar * P.lift(self))

        _lmul_ = _acted_upon_
        _rmul_ = _acted_upon_


class FiniteDimensionalSimpleModule(SimpleModule):
    """
    A finite dimensional simple module.
    """
    def bgg_resolution(self):
        """
        Return the BGG resolution of ``self``.

        EXAMPLES::

            sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
            sage: La = g.cartan_type().root_system().weight_lattice().fundamental_weights()
            sage: L = g.simple_module(La[1] + La[2])
            sage: L.bgg_resolution()
            BGG resolution of Simple module with highest weight Lambda[1] + Lambda[2]
             of Lie algebra of ['A', 2] in the Chevalley basis
        """
        from sage.algebras.lie_algebras.bgg_resolution import BGGResolution
        return BGGResolution(self)



# ***************************************************************************
#       Copyright (C) 2017 Jeroen Demeyer <J.Demeyer@UGent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

cdef extern from "ccobject.h":
    # Print representation of any C++ object
    str ccrepr[T](const T& x)

    # Read a Python bytes/str into a C++ object
    int ccreadstr[T](T x, object b) except -1



# sage.doctest: optional - numpy
"""
C Int Lists

This is a class for fast basic operations with lists of C ints.  It is
similar to the double precision TimeSeries class.  It has all the
standard C int semantics, of course, including overflow.  It is also
similar to the Python list class, except all elements are C ints,
which makes some operations much, much faster.  For example,
concatenating two IntLists can be over 10 times faster than
concatenating the corresponding Python lists of ints, and taking
slices is also much faster.

AUTHOR:

- William Stein, 2010-03
"""

# ****************************************************************************
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

# Global parameter that sets the maximum number of entries of an IntList to print.
max_print = 10

cimport cython
from libc.string cimport memcpy
from cysignals.memory cimport sig_malloc, sig_free
from cysignals.signals cimport sig_on, sig_off

from sage.rings.integer import Integer
from sage.stats.time_series cimport TimeSeries
from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AsString
from sage.structure.richcmp cimport rich_to_bool


cdef class IntList:
    """
    A list of C int's.
    """
    def __cinit__(self):
        """
        Create new empty uninitialized IntList.

        EXAMPLES::

            sage: stats.IntList(5)     # indirect test
            [0, 0, 0, 0, 0]
        """
        self._values = NULL

    def __init__(self, values):
        """
        Create an initialized list of C ints.

        INPUT:

        - ``values`` -- int, long, Integer, list of integers, or a TimeSeries

        If the input is a time series or list of floats, then the
        integer parts of the entries are taken (not the floor).

        EXAMPLES::

            sage: stats.IntList(8)
            [0, 0, 0, 0, 0, 0, 0, 0]

            sage: stats.IntList([1,5,-39392])
            [1, 5, -39392]

        We check for overflow when creating the IntList::

            sage: stats.IntList([1, 3, 2^32])
            Traceback (most recent call last):
            ...
            OverflowError: ... too large to convert to ...

        Printing omits entries::

            sage: stats.IntList(1000)
            [0, 0, 0, 0, 0 ... 0, 0, 0, 0, 0]

        Floats are truncated to their integer parts::

            sage: stats.IntList([1.1, -2.6])
            [1, -2]
            sage: stats.IntList(stats.TimeSeries([1.1, -2.6]))
            [1, -2]
        """
        cdef TimeSeries T
        if isinstance(values, (int, Integer)):
            self._length = values
            values = None
        elif isinstance(values, TimeSeries):
            T = values
            self._length = T._length
        else:
            self._length = len(values)

        self._values = <int*> sig_malloc(sizeof(int)*self._length)
        if self._values == NULL:
            raise MemoryError
        cdef Py_ssize_t i
        if values is None:
            for i in range(self._length):
                self._values[i] = 0
        elif isinstance(values, TimeSeries):
            for i in range(self._length):
                self._values[i] = <int> T._values[i]
        else:
            for i in range(self._length):
                self._values[i] = values[i]

    def __richcmp__(IntList self, other, int op):
        """
        Compare ``self`` and ``other``.  This has the same semantics
        as list comparison.

        EXAMPLES::

            sage: v = stats.IntList([1,2,3]); w = stats.IntList([1,2])
            sage: v < w
            False
            sage: w < v
            True
            sage: v == v
            True
            sage: w == w
            True
        """
        cdef IntList _other
        cdef Py_ssize_t c, i
        if not isinstance(other, IntList):
            _other = IntList(other)
        else:
            _other = <IntList>other
        for i in range(min(self._length, _other._length)):
            d = self._values[i] - _other._values[i]
            if d:
                return rich_to_bool(op, -1 if d < 0 else 1)
        c = self._length - _other._length
        if c:
            return rich_to_bool(op, -1 if c < 0 else 1)
        return rich_to_bool(op, 0)

    def __dealloc__(self):
        """
        Deallocate memory used by the IntList, if it was allocated.
        """
        sig_free(self._values)

    def __repr__(self):
        """
        Return string representation of this IntList.

        EXAMPLES::

            sage: a = stats.IntList([1..15]); a.__repr__()
            '[1, 2, 3, 4, 5 ... 11, 12, 13, 14, 15]'
            sage: sage.stats.intlist.max_print = 20
            sage: a.__repr__()
            '[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]'
            sage: sage.stats.intlist.max_print = 10
            sage: a.__repr__()
            '[1, 2, 3, 4, 5 ... 11, 12, 13, 14, 15]'
        """
        if len(self) > max_print:
            v0 = self[:max_print//2]
            v1 = self[-max_print//2:]
            return '[' + ', '.join(str(x) for x in v0) + ' ... ' + \
                         ', '.join(str(x) for x in v1) + ']'
        else:
            return str(self.list())

    def __getitem__(self, i):
        """
        Return i-th entry or slice of self, following standard Python
        semantics.  The returned slice is an intlist, and the returned
        entry is a Python int.

        INPUT:

        - ``i`` -- integer or slice

        EXAMPLES::

            sage: a = stats.IntList([0..9]); a
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            sage: a[5]
            5
            sage: a[-2]
            8
            sage: a[5:-2]
            [5, 6, 7]
            sage: type(a[5:-2])
            <... 'sage.stats.intlist.IntList'>
            sage: type(a[5])
            <... 'int'>
        """
        cdef Py_ssize_t start, stop, step, j
        cdef IntList t
        if isinstance(i, slice):
            start = 0 if (i.start is None) else i.start
            stop = self._length if (i.stop is None) else i.stop
            step = 1 if (i.step is None) else i.step
            if start < 0:
                start += self._length
                if start < 0: start = 0
            elif start >= self._length:
                start = self._length - 1
            if stop < 0:
                stop += self._length
                if stop < 0: stop = 0
            elif stop > self._length:
                stop = self._length
            if start >= stop:
                return new_int_list(0)
            if step < 0:
                step = -step
                t = new_int_list((stop-start)/step)
                for j from 0 <= j < (stop-start)/step:
                    t._values[j] = self._values[stop-1 - j*step]
            elif step > 1:
                t = new_int_list((stop-start)/step)
                for j from 0 <= j < (stop-start)/step:
                    t._values[j] = self._values[j*step+start]
            else:
                t = new_int_list(stop-start)
                memcpy(t._values, self._values + start, sizeof(int)*t._length)
            return t
        else:
            j = i
            if j < 0:
                j += self._length
                if j < 0:
                    raise IndexError("IntList index out of range")
            elif j >= self._length:
                raise IndexError("IntList index out of range")
            return self._values[j]

    def __setitem__(self, Py_ssize_t i, int x):
        """
        Set the i-th entry of self, following standard Python semantics.

        INPUT:

            - ``i`` -- integer
            - ``x`` -- integer

        EXAMPLES::

            sage: a = stats.IntList([-2,3,7,-4])
            sage: a[1] = 10393; a
            [-2, 10393, 7, -4]
            sage: a[-1] = -10; a
            [-2, 10393, 7, -10]
            sage: a[100]
            Traceback (most recent call last):
            ...
            IndexError: IntList index out of range
            sage: a[-100]
            Traceback (most recent call last):
            ...
            IndexError: IntList index out of range
        """
        if i < 0:
            i += self._length
            if i < 0:
                raise IndexError("index out of range")
        elif i >= self._length:
            raise IndexError("index out of range")
        self._values[i] = x

    def __reduce__(self):
        """
        Used in pickling int lists.

        EXAMPLES::

            sage: a = stats.IntList([-2,3,7,-4])
            sage: loads(dumps(a)) == a
            True

            sage: v = stats.IntList([1,-3])
            sage: v.__reduce__()
            (<cyfunction unpickle_intlist_v1 at ...>, (..., 2))
            sage: loads(dumps(v)) == v
            True

        Note that dumping and loading with compress ``False`` is much faster,
        though dumping with compress ``True`` can save a lot of space::

            sage: v = stats.IntList([1..10^5])
            sage: loads(dumps(v, compress=False),compress=False) == v
            True
        """
        buf = PyBytes_FromStringAndSize(<char*>self._values, self._length*sizeof(int)/sizeof(char))
        return unpickle_intlist_v1, (buf, self._length)

    def list(self):
        """
        Return Python list version of ``self`` with Python ints as entries.

        EXAMPLES::

            sage: a = stats.IntList([1..15]); a
            [1, 2, 3, 4, 5 ... 11, 12, 13, 14, 15]
            sage: a.list()
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
            sage: list(a) == a.list()
            True
            sage: type(a.list()[0])
            <... 'int'>
        """
        cdef Py_ssize_t i
        return [self._values[i] for i in range(self._length)]

    cpdef int sum(self) noexcept:
        """
        Return the sum of the entries of ``self``.

        EXAMPLES::

            sage: stats.IntList([1..100]).sum()
            5050

        Note that there can be overflow, since the entries are C ints::

            sage: a = stats.IntList([2^30,2^30]); a
            [1073741824, 1073741824]
            sage: a.sum()
            -2147483648
        """
        cdef Py_ssize_t i
        cdef int s=0
        sig_on()
        for i in range(self._length):
            s += self._values[i]
        sig_off()
        return s

    cpdef int prod(self) noexcept:
        """
        Return the product of the entries of ``self``.

        EXAMPLES::

            sage: a = stats.IntList([1..10]); a
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            sage: a.prod()
            3628800
            sage: factorial(10)
            3628800

        Note that there can be overflow::

            sage: a = stats.IntList([2^30, 2]); a
            [1073741824, 2]
            sage: a.prod()
            -2147483648
         """
        cdef Py_ssize_t i
        cdef int s=1
        sig_on()
        for i in range(self._length):
            s *= self._values[i]
        sig_off()
        return s

    def __len__(self):
        """
        Return the number of entries in this time series.

        OUTPUT: Python integer

        EXAMPLES::

            sage: len(stats.IntList([1..15]))
            15
            sage: len(stats.IntList([]))
            0
            sage: len(stats.IntList(10^6))
            1000000
        """
        return self._length

    def __add__(left, right):
        """
        Concatenate the integer lists ``self`` and ``right``.

        EXAMPLES::

            sage: stats.IntList([-2,3,5]) + stats.IntList([1,1,17])
            [-2, 3, 5, 1, 1, 17]
        """
        if not isinstance(right, IntList):
            raise TypeError("right operand must be an int list")
        if not isinstance(left, IntList):
            raise TypeError("left operand must be an int list")
        cdef IntList R = right
        cdef IntList L = left
        cdef IntList t = new_int_list(L._length + R._length)
        memcpy(t._values, L._values, sizeof(int)*L._length)
        memcpy(t._values + L._length, R._values, sizeof(int)*R._length)
        return t

    def min(self, bint index=False):
        """
        Return the smallest value in this integer list.  If this
        series has length 0 we raise a :exc:`ValueError`.

        INPUT:

        - ``index`` -- boolean (default: ``False``); if ``True``, also return
          index of minimal entry

        OUTPUT:

        - ``float`` -- smallest value
        - ``integer`` -- index of smallest value; only returned if
          ``index=True``

        EXAMPLES::

            sage: v = stats.IntList([1,-4,3,-2,-4])
            sage: v.min()
            -4
            sage: v.min(index=True)
            (-4, 1)
        """
        if self._length == 0:
            raise ValueError("min() arg is an empty sequence")
        cdef Py_ssize_t i, j
        cdef int s = self._values[0]
        j = 0
        for i in range(1, self._length):
            if self._values[i] < s:
                s = self._values[i]
                j = i
        if index:
            return s, j
        else:
            return s

    def max(self, bint index=False):
        """
        Return the largest value in this time series. If this series
        has length 0 we raise a :exc:`ValueError`

        INPUT:

        - ``index`` -- boolean (default: ``False``); if ``True``, also return
          index of maximum entry

        OUTPUT:

        - integer -- largest value
        - integer -- index of largest value; only returned if ``index=True``

        EXAMPLES::

            sage: v = stats.IntList([1,-4,3,-2,-4,3])
            sage: v.max()
            3
            sage: v.max(index=True)
            (3, 2)
        """
        if self._length == 0:
            raise ValueError("max() arg is an empty sequence")
        cdef Py_ssize_t i, j = 0
        cdef int s = self._values[0]
        for i in range(1,self._length):
            if self._values[i] > s:
                s = self._values[i]
                j = i
        if index:
            return s, j
        else:
            return s

    def time_series(self):
        """
        Return :class:`TimeSeries` version of ``self``, which involves changing
        each entry to a double.

        EXAMPLES::

            sage: T = stats.IntList([-2,3,5]).time_series(); T
            [-2.0000, 3.0000, 5.0000]
            sage: type(T)
            <... 'sage.stats.time_series.TimeSeries'>
        """
        cdef TimeSeries T = TimeSeries.__new__(TimeSeries)
        # We just reach into the data structure underlying T, since we
        # want this function to be *very* fast.
        T._length = self._length
        T._values = <double*> sig_malloc(sizeof(double)*self._length)
        cdef Py_ssize_t i
        for i in range(self._length):
            T._values[i] = self._values[i]
        return T

    def plot(self, *args, **kwds):
        """
        Return a plot of this :class:`IntList`.

        This just constructs the
        corresponding double-precision floating point :class:`TimeSeries`
        object, passing on all arguments.

        EXAMPLES::

            sage: stats.IntList([3,7,19,-2]).plot()                                     # needs sage.plot
            Graphics object consisting of 1 graphics primitive
            sage: stats.IntList([3,7,19,-2]).plot(color='red',                          # needs sage.plot
            ....:                                 pointsize=50, points=True)
            Graphics object consisting of 1 graphics primitive
        """
        return self.time_series().plot(*args, **kwds)

    def plot_histogram(self, *args, **kwds):
        """
        Return a histogram plot of this :class:`IntList`.

        This just constructs
        the corresponding double-precision floating point :class:`TimeSeries` object,
        and plots it, passing on all arguments.

        EXAMPLES::

            sage: stats.IntList([1..15]).plot_histogram()                               # needs sage.plot
            Graphics object consisting of 50 graphics primitives
        """
        return self.time_series().plot_histogram(*args, **kwds)


cdef IntList new_int_list(Py_ssize_t length):
    """
    Function that is used internally to quickly create a new intlist
    without initializing any of the allocated memory.

    INPUT:

    - ``length`` -- nonnegative integer

    OUTPUT: an IntList
    """
    if length < 0:
        raise ValueError("length must be nonnegative")
    cdef IntList t = IntList.__new__(IntList)
    t._length = length
    t._values = <int*> sig_malloc(sizeof(int)*length)
    return t


def unpickle_intlist_v1(bytes v, Py_ssize_t n):
    """
    Version 1 unpickle method.

    INPUT:

    - ``v`` -- a raw char buffer

    EXAMPLES::

        sage: v = stats.IntList([1,2,3])
        sage: s = v.__reduce__()[1][0]
        sage: type(s) == type(b'')
        True
        sage: sage.stats.intlist.unpickle_intlist_v1(s, 3)
        [1, 2, 3]
        sage: sage.stats.intlist.unpickle_intlist_v1(s+s,6)
        [1, 2, 3, 1, 2, 3]
        sage: sage.stats.intlist.unpickle_intlist_v1(b'',0)
        []
    """
    cdef IntList t = new_int_list(n)
    memcpy(t._values, PyBytes_AsString(v), n*sizeof(int))
    return t



"""
Differential Geometry of Parametrized Surfaces

AUTHORS:

- Mikhail Malakhaltsev (2010-09-25): initial version
- Joris Vankerschaver  (2010-10-25): implementation, doctests
"""
# ****************************************************************************
#       Copyright (C) 2010  Mikhail Malakhaltsev <mikarm@gmail.com>
#       Copyright (C) 2010  Joris Vankerschaver <joris.vankerschaver@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from itertools import product

from sage.structure.sage_object import SageObject
from sage.modules.free_module_element import vector
from sage.matrix.constructor import matrix
from sage.calculus.functional import diff
from sage.misc.functional import sqrt
from sage.misc.cachefunc import cached_method
from sage.symbolic.ring import SR
from sage.symbolic.constants import pi


def _simplify_full_rad(f):
    """
    Helper function to conveniently call :meth:`simplify_full` and
    :meth:`canonicalize_radical` in succession.

    INPUT:

    - ``f`` -- a symbolic expression

    EXAMPLES::

        sage: from sage.geometry.riemannian_manifolds.parametrized_surface3d import _simplify_full_rad
        sage: _simplify_full_rad(sqrt(x^2)/x)
        1
    """
    return f.simplify_full().canonicalize_radical()


class ParametrizedSurface3D(SageObject):
    r"""
    Class representing a parametrized two-dimensional surface in
    Euclidian three-space.  Provides methods for calculating the main
    geometrical objects related to such a surface, such as the first
    and the second fundamental form, the total (Gaussian) and the mean
    curvature, the geodesic curves, parallel transport, etc.


    INPUT:

    - ``surface_equation`` -- a 3-tuple of functions specifying a parametric
      representation of the surface

    - ``variables`` -- a 2-tuple of intrinsic coordinates `(u, v)` on the
      surface, with `u` and `v` symbolic variables, or a 2-tuple of triples
      `(u, u_{min}, u_{max})`, `(v, v_{min}, v_{max})` when the parameter range
      for the coordinates is known

    - ``name`` -- name of the surface (optional)


    .. NOTE::

       Throughout the documentation, we use the Einstein summation
       convention: whenever an index appears twice, once as a
       subscript, and once as a superscript, summation over that index
       is implied.  For instance, `g_{ij} g^{jk}` stands for `\sum_j
       g_{ij}g^{jk}`.

    EXAMPLES:

    We give several examples of standard surfaces in differential
    geometry.  First, let's construct an elliptic paraboloid by
    explicitly specifying its parametric equation::

        sage: u, v = var('u,v', domain='real')
        sage: eparaboloid = ParametrizedSurface3D((u, v, u^2 + v^2), (u, v),
        ....:                                     'elliptic paraboloid'); eparaboloid
        Parametrized surface ('elliptic paraboloid') with equation (u, v, u^2 + v^2)

    When the ranges for the intrinsic coordinates are known, they can be
    specified explicitly. This is mainly useful for plotting. Here we
    construct half of an ellipsoid::

        sage: u1, u2 = var ('u1, u2', domain='real')
        sage: coords = ((u1, -pi/2, pi/2), (u2, 0, pi))
        sage: ellipsoid_eq = (cos(u1)*cos(u2), 2*sin(u1)*cos(u2), 3*sin(u2))
        sage: ellipsoid = ParametrizedSurface3D(ellipsoid_eq, coords, 'ellipsoid'); ellipsoid
        Parametrized surface ('ellipsoid') with equation
        (cos(u1)*cos(u2), 2*cos(u2)*sin(u1), 3*sin(u2))
        sage: ellipsoid.plot()                                                          # needs sage.plot
        Graphics3d Object

    Standard surfaces can be constructed using the ``surfaces`` generator::

        sage: klein = surfaces.Klein(); klein
        Parametrized surface ('Klein bottle') with equation
        (-(sin(1/2*u)*sin(2*v) - cos(1/2*u)*sin(v) - 1)*cos(u),
         -(sin(1/2*u)*sin(2*v) - cos(1/2*u)*sin(v) - 1)*sin(u),
         cos(1/2*u)*sin(2*v) + sin(1/2*u)*sin(v))

    Latex representation of the surfaces::

        sage: u, v = var('u, v', domain='real')
        sage: sphere = ParametrizedSurface3D((cos(u)*cos(v), sin(u)*cos(v), sin(v)), (u, v),
        ....:                                'sphere')
        sage: print(latex(sphere))
        \left(\cos\left(u\right) \cos\left(v\right), \cos\left(v\right) \sin\left(u\right), \sin\left(v\right)\right)
        sage: print(sphere._latex_())
        \left(\cos\left(u\right) \cos\left(v\right), \cos\left(v\right) \sin\left(u\right), \sin\left(v\right)\right)
        sage: print(sphere)
        Parametrized surface ('sphere') with equation (cos(u)*cos(v), cos(v)*sin(u), sin(v))

    To plot a parametric surface, use the :meth:`plot` member function::

        sage: enneper = surfaces.Enneper(); enneper
        Parametrized surface ('Enneper's surface') with equation
        (-1/9*(u^2 - 3*v^2 - 3)*u, -1/9*(3*u^2 - v^2 + 3)*v, 1/3*u^2 - 1/3*v^2)
        sage: enneper.plot(aspect_ratio='automatic')                                    # needs sage.plot
        Graphics3d Object

    We construct an ellipsoid whose axes are given by symbolic variables `a`,
    `b` and `c`, and find the natural frame of tangent vectors,
    expressed in intrinsic coordinates. Note that the result is a
    dictionary of vector fields::

        sage: a, b, c = var('a, b, c', domain='real')
        sage: u1, u2 = var('u1, u2', domain='real')
        sage: ellipsoid_eq = (a*cos(u1)*cos(u2), b*sin(u1)*cos(u2), c*sin(u2))
        sage: ellipsoid = ParametrizedSurface3D(ellipsoid_eq, (u1, u2),
        ....:                                   'Symbolic ellipsoid'); ellipsoid
        Parametrized surface ('Symbolic ellipsoid') with equation
        (a*cos(u1)*cos(u2), b*cos(u2)*sin(u1), c*sin(u2))

        sage: ellipsoid.natural_frame()
        {1: (-a*cos(u2)*sin(u1), b*cos(u1)*cos(u2), 0),
         2: (-a*cos(u1)*sin(u2), -b*sin(u1)*sin(u2), c*cos(u2))}

    We find the normal vector field to the surface.  The normal vector
    field is the vector product of the vectors of the natural frame,
    and is given by::

        sage: ellipsoid.normal_vector()
        (b*c*cos(u1)*cos(u2)^2, a*c*cos(u2)^2*sin(u1), a*b*cos(u2)*sin(u2))

    By default, the normal vector field is not normalized.  To obtain
    the unit normal vector field of the elliptic paraboloid, we put::

        sage: u, v = var('u,v', domain='real')
        sage: eparaboloid = ParametrizedSurface3D([u, v, u^2 + v^2], [u,v],
        ....:                                     'elliptic paraboloid')
        sage: eparaboloid.normal_vector(normalized=True)
        (-2*u/sqrt(4*u^2 + 4*v^2 + 1), -2*v/sqrt(4*u^2 + 4*v^2 + 1), 1/sqrt(4*u^2 + 4*v^2 + 1))

    Now let us compute the coefficients of the first fundamental form of the torus::

        sage: u, v = var('u, v', domain='real')
        sage: a, b = var('a, b', domain='real')
        sage: torus = ParametrizedSurface3D(((a + b*cos(u))*cos(v),
        ....:                                (a + b*cos(u))*sin(v),
        ....:                                b*sin(u)), [u,v], 'torus')
        sage: torus.first_fundamental_form_coefficients()
        {(1, 1): b^2, (1, 2): 0, (2, 1): 0, (2, 2): b^2*cos(u)^2 + 2*a*b*cos(u) + a^2}

    The first fundamental form can be used to compute the length of a
    curve on the surface.  For example, let us find the length of the
    curve `u^1 = t`, `u^2 = t`, `t \in [0,2\pi]`, on the ellipsoid
    with axes `a=1`, `b=1.5` and `c=1`. So we take the curve::

        sage: t = var('t', domain='real')
        sage: u1 = t
        sage: u2 = t

    Then find the tangent vector::

        sage: du1 = diff(u1,t)
        sage: du2 = diff(u2,t)
        sage: du = vector([du1, du2]); du
        (1, 1)

    Once we specify numerical values for the axes of the ellipsoid, we can
    determine the numerical value of the length integral::

        sage: L = sqrt(ellipsoid.first_fundamental_form(du, du).substitute(u1=u1, u2=u2))
        sage: numerical_integral(L.substitute(a=2, b=1.5, c=1),0,1)[0] # rel tol 1e-11
        2.00127905972

    We find the area of the sphere of radius `R`::

        sage: R = var('R', domain='real')
        sage: u, v = var('u,v', domain='real')
        sage: assume(R>0)
        sage: assume(cos(v)>0)
        sage: sphere = ParametrizedSurface3D([R*cos(u)*cos(v), R*sin(u)*cos(v), R*sin(v)],
        ....:                                [u,v], 'sphere')
        sage: integral(integral(sphere.area_form(), u, 0, 2*pi), v, -pi/2, pi/2)
        4*pi*R^2

    We can find an orthonormal frame field `\{e_1, e_2\}` of a surface
    and calculate its structure functions.  Let us first determine the
    orthonormal frame field for the elliptic paraboloid::

        sage: u, v = var('u,v', domain='real')
        sage: eparaboloid = ParametrizedSurface3D([u, v, u^2 + v^2], [u,v],
        ....:                                     'elliptic paraboloid')
        sage: eparaboloid.orthonormal_frame()
        {1: (1/sqrt(4*u^2 + 1), 0, 2*u/sqrt(4*u^2 + 1)),
         2: (-4*u*v/(sqrt(4*u^2 + 4*v^2 + 1)*sqrt(4*u^2 + 1)),
             sqrt(4*u^2 + 1)/sqrt(4*u^2 + 4*v^2 + 1),
             2*v/(sqrt(4*u^2 + 4*v^2 + 1)*sqrt(4*u^2 + 1)))}

    We can express the orthogonal frame field both in exterior
    coordinates (i.e. expressed as vector field fields in the ambient
    space `\RR^3`, the default) or in intrinsic coordinates
    (with respect to the natural frame).  Here we use intrinsic
    coordinates::

        sage: eparaboloid.orthonormal_frame(coordinates='int')
        {1: (1/sqrt(4*u^2 + 1), 0),
         2: (-4*u*v/(sqrt(4*u^2 + 4*v^2 + 1)*sqrt(4*u^2 + 1)),
             sqrt(4*u^2 + 1)/sqrt(4*u^2 + 4*v^2 + 1))}

    Using the orthonormal frame in interior coordinates, we can calculate
    the structure functions `c^k_{ij}` of the surface, defined by
    `[e_i,e_j] =  c^k_{ij} e_k`, where `[e_i, e_j]` represents the Lie
    bracket of two frame vector fields `e_i, e_j`.  For the
    elliptic paraboloid, we get::

        sage: EE = eparaboloid.orthonormal_frame(coordinates='int')
        sage: E1 = EE[1]; E2 = EE[2]
        sage: CC = eparaboloid.frame_structure_functions(E1,E2)
        sage: CC[1,2,1].simplify_full()
        4*sqrt(4*u^2 + 4*v^2 + 1)*v/((16*u^4 + 4*(4*u^2 + 1)*v^2 + 8*u^2 + 1)*sqrt(4*u^2 + 1))

    We compute the Gaussian and mean curvatures of the sphere::

        sage: sphere = surfaces.Sphere(); sphere
        Parametrized surface ('Sphere') with equation (cos(u)*cos(v), cos(v)*sin(u), sin(v))
        sage: K = sphere.gauss_curvature(); K  # Not tested -- see trac 12737
        1
        sage: H = sphere.mean_curvature(); H  # Not tested -- see trac 12737
        -1

    We can easily generate a color plot of the Gaussian curvature of a surface.
    Here we deal with the ellipsoid::

        sage: # needs numpy
        sage: u1, u2 = var('u1,u2', domain='real')
        sage: u = [u1,u2]
        sage: ellipsoid_equation(u1,u2) = [2*cos(u1)*cos(u2),1.5*cos(u1)*sin(u2),sin(u1)]
        sage: ellipsoid = ParametrizedSurface3D(ellipsoid_equation(u1,u2), [u1, u2], 'ellipsoid')
        sage: # set intervals for variables and the number of division points
        sage: u1min, u1max = -1.5, 1.5
        sage: u2min, u2max = 0, 6.28
        sage: u1num, u2num = 10, 20
        sage: # make the arguments array
        sage: from numpy import linspace
        sage: u1_array = linspace(u1min, u1max, u1num)
        sage: u2_array = linspace(u2min, u2max, u2num)
        sage: u_array = [(uu1,uu2) for uu1 in u1_array for uu2 in u2_array]
        sage: # Find the gaussian curvature
        sage: K(u1,u2) = ellipsoid.gauss_curvature()
        sage: # Make array of K values
        sage: K_array = [K(uu[0],uu[1]) for uu in u_array]
        sage: # Find minimum and max of the Gauss curvature
        sage: K_max = max(K_array)
        sage: K_min = min(K_array)
        sage: # Make the array of color coefficients
        sage: cc_array = [(ccc - K_min)/(K_max - K_min) for ccc in K_array]
        sage: points_array = [ellipsoid_equation(u_array[counter][0],
        ....:                                    u_array[counter][1])
        ....:                 for counter in range(len(u_array))]
        sage: curvature_ellipsoid_plot = sum(point([xx                                  # needs sage.plot
        ....:                                       for xx in points_array[counter]],
        ....:                                      color=hue(cc_array[counter]/2))
        ....:                                for counter in range(len(u_array)))
        sage: curvature_ellipsoid_plot.show(aspect_ratio=1)                             # needs sage.plot

    We can find the principal curvatures and principal directions of the
    elliptic paraboloid::

        sage: u, v = var('u, v', domain='real')
        sage: eparaboloid = ParametrizedSurface3D([u, v, u^2+v^2], [u, v],
        ....:                                     'elliptic paraboloid')
        sage: pd = eparaboloid.principal_directions(); pd
        [(2*sqrt(4*u^2 + 4*v^2 + 1)/(16*u^4 + 16*v^4 + 8*(4*u^2 + 1)*v^2 + 8*u^2 + 1), [(1, v/u)], 1),
         (2/sqrt(4*u^2 + 4*v^2 + 1), [(1, -u/v)], 1)]

    We extract the principal curvatures::

        sage: k1 = pd[0][0].simplify_full(); k1
        2*sqrt(4*u^2 + 4*v^2 + 1)/(16*u^4 + 16*v^4 + 8*(4*u^2 + 1)*v^2 + 8*u^2 + 1)
        sage: k2 = pd[1][0].simplify_full(); k2
        2/sqrt(4*u^2 + 4*v^2 + 1)

    and check them by comparison with the Gaussian and mean curvature
    expressed in terms of the principal curvatures::

        sage: K = eparaboloid.gauss_curvature().simplify_full(); K
        4/(16*u^4 + 16*v^4 + 8*(4*u^2 + 1)*v^2 + 8*u^2 + 1)
        sage: H = eparaboloid.mean_curvature().simplify_full(); H
        2*(2*u^2 + 2*v^2 + 1)/(4*u^2 + 4*v^2 + 1)^(3/2)
        sage: (K - k1*k2).simplify_full()
        0
        sage: (2*H - k1 - k2).simplify_full()
        0

    We can find the intrinsic (local coordinates) of the principal directions::

        sage: pd[0][1]
        [(1, v/u)]
        sage: pd[1][1]
        [(1, -u/v)]

    The ParametrizedSurface3D class also contains functionality to
    compute the coefficients of the second fundamental form, the shape
    operator, the rotation on the surface at a given angle, the
    connection coefficients.  One can also calculate numerically the
    geodesics and the parallel translation along a curve.

    Here we compute a number of geodesics on the sphere emanating
    from the point ``(1, 0, 0)``, in various directions. The geodesics
    intersect again in the antipodal point ``(-1, 0, 0)``, indicating
    that these points are conjugate::

        sage: S = surfaces.Sphere()
        sage: g1 = [c[-1] for c in S.geodesics_numerical((0,0), (1,0), (0,2*pi,100))]
        sage: g2 = [c[-1] for c in S.geodesics_numerical((0,0),
        ....:                                            (cos(pi/3),sin(pi/3)),
        ....:                                            (0,2*pi,100))]
        sage: g3 = [c[-1] for c in S.geodesics_numerical((0,0),
        ....:                                            (cos(2*pi/3),sin(2*pi/3)),
        ....:                                            (0,2*pi,100))]
        sage: (S.plot(opacity=0.3) + line3d(g1, color='red')                            # needs sage.plot
        ....:     + line3d(g2, color='red') + line3d(g3, color='red')).show()
    """

    def __init__(self, equation, variables, name=None):
        r"""
        See ``ParametrizedSurface3D`` for full documentation.

        .. NOTE::

            The orientation of the surface is determined by the
            parametrization, that is, the natural frame with positive
            orientation is given by `\partial_1 \vec r`, `\partial_2 \vec
            r`.

        EXAMPLES::

            sage: u, v = var('u,v', domain='real')
            sage: eq = (3*u + 3*u*v^2 - u^3, 3*v + 3*u^2*v - v^3, 3*(u^2-v^2))
            sage: enneper = ParametrizedSurface3D(eq, (u, v),'Enneper Surface'); enneper
            Parametrized surface ('Enneper Surface') with equation
            (-u^3 + 3*u*v^2 + 3*u, 3*u^2*v - v^3 + 3*v, 3*u^2 - 3*v^2)
        """
        self.equation = tuple(equation)

        if isinstance(variables[0], (list, tuple)):
            self.variables_range = (variables[0][1:3], variables[1][1:3])
            self.variables_list = (variables[0][0], variables[1][0])
        else:
            self.variables_range = None
            self.variables_list = variables

        self.variables = {1:self.variables_list[0], 2:self.variables_list[1]}
        self.name = name

    def _latex_(self):
        r"""
        Return the LaTeX representation of this parametrized surface.

        EXAMPLES::

            sage: u, v = var('u, v')
            sage: sphere = ParametrizedSurface3D((cos(u)*cos(v), sin(u)*cos(v), sin(v)), (u, v),'sphere')
            sage: latex(sphere)
            \left(\cos\left(u\right) \cos\left(v\right), \cos\left(v\right) \sin\left(u\right), \sin\left(v\right)\right)
            sage: sphere._latex_()
            \left(\cos\left(u\right) \cos\left(v\right), \cos\left(v\right) \sin\left(u\right), \sin\left(v\right)\right)
        """
        from sage.misc.latex import latex
        return latex(self.equation)

    def _repr_(self):
        r"""
        Return the string representation of this parametrized surface.

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: eq = (3*u + 3*u*v^2 - u^3, 3*v + 3*u^2*v - v^3, 3*(u^2-v^2))
            sage: enneper = ParametrizedSurface3D(eq,[u,v],'enneper_surface')
            sage: print(enneper)
            Parametrized surface ('enneper_surface') with equation (-u^3 + 3*u*v^2 + 3*u, 3*u^2*v - v^3 + 3*v, 3*u^2 - 3*v^2)
            sage: enneper._repr_()
            "Parametrized surface ('enneper_surface') with equation (-u^3 + 3*u*v^2 + 3*u, 3*u^2*v - v^3 + 3*v, 3*u^2 - 3*v^2)"
        """
        name = 'Parametrized surface'
        if self.name is not None:
            name += " ('%s')" % self.name
        s = '%(designation)s with equation %(eq)s' % \
            {'designation': name, 'eq': str(self.equation)}
        return s

    def point(self, coords):
        r"""
        Return a point on the surface given its intrinsic coordinates.

        INPUT:

        - ``coords`` -- 2-tuple specifying the intrinsic coordinates ``(u, v)`` of the point

        OUTPUT: 3-vector specifying the coordinates in `\RR^3` of the point

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: torus = ParametrizedSurface3D(((2 + cos(u))*cos(v),
            ....:                                (2 + cos(u))*sin(v),
            ....:                                sin(u)), [u,v], 'torus')
            sage: torus.point((0, pi/2))
            (0, 3, 0)
            sage: torus.point((pi/2, pi))
            (-2, 0, 1)
            sage: torus.point((pi, pi/2))
            (0, 1, 0)
        """

        d = dict(zip(self.variables_list, coords))
        return vector([f.subs(d) for f in self.equation])

    def tangent_vector(self, coords, components):
        r"""
        Return the components of a tangent vector given the intrinsic
        coordinates of the base point and the components of the vector
        in the intrinsic frame.

        INPUT:

        - ``coords`` -- 2-tuple specifying the intrinsic coordinates ``(u, v)`` of the point

        - ``components`` -- 2-tuple specifying the components of the tangent
          vector in the intrinsic coordinate frame

        OUTPUT: 3-vector specifying the components in `\RR^3` of the vector

        EXAMPLES:

        We compute two tangent vectors to Enneper's surface along the
        coordinate lines and check that their cross product gives the
        normal vector::

            sage: u, v = var('u,v', domain='real')
            sage: eq = (3*u + 3*u*v^2 - u^3, 3*v + 3*u^2*v - v^3, 3*(u^2-v^2))
            sage: e = ParametrizedSurface3D(eq, (u, v), 'Enneper Surface')

            sage: w1 = e.tangent_vector((1, 2), (1, 0)); w1
            (12, 12, 6)
            sage: w2 = e.tangent_vector((1, 2), (0, 1)); w2
            (12, -6, -12)
            sage: w1.cross_product(w2)
            (-108, 216, -216)

            sage: n = e.normal_vector().subs({u: 1, v: 2}); n
            (-108, 216, -216)
            sage: n == w1.cross_product(w2)
            True
        """

        components = vector(components)
        d = dict(zip(self.variables_list, coords))
        jacobian = matrix([[f.diff(u).subs(d) for u in self.variables_list]
                           for f in self.equation])
        return jacobian * components

    def plot(self, urange=None, vrange=None, **kwds):
        r"""
        Enable easy plotting directly from the surface class.

        The optional keywords ``urange`` and ``vrange`` specify the range for
        the surface parameters `u` and `v`. If either of these parameters
        is ``None``, the method checks whether a parameter range was
        specified when the surface was created. If not, the default of
        `(0, 2 \pi)` is used.

        INPUT:

        - ``urange`` -- 2-tuple specifying the parameter range for `u`
        - ``vrange`` -- 2-tuple specifying the parameter range for `v`

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: eq = (3*u + 3*u*v^2 - u^3, 3*v + 3*u^2*v - v^3, 3*(u^2-v^2))
            sage: enneper = ParametrizedSurface3D(eq, (u, v), 'Enneper Surface')
            sage: enneper.plot((-5, 5), (-5, 5))                                        # needs sage.plot
            Graphics3d Object
        """

        from sage.plot.plot3d.parametric_plot3d import parametric_plot3d

        if self.variables_range is None:
            if urange is None:
                urange = (0, 2*pi)
            if vrange is None:
                vrange = (0, 2*pi)
        else:
            if urange is None:
                urange = self.variables_range[0]
            if vrange is None:
                vrange = self.variables_range[1]

        urange3 = (self.variables[1],) + tuple(urange)
        vrange3 = (self.variables[2],) + tuple(vrange)
        P = parametric_plot3d(self.equation, urange3, vrange3, **kwds)

        return P

    @cached_method
    def natural_frame(self):
        """
        Return the natural tangent frame on the parametrized surface.
        The vectors of this frame are tangent to the coordinate lines
        on the surface.

        OUTPUT: the natural frame as a dictionary

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: eparaboloid = ParametrizedSurface3D((u, v, u^2+v^2), (u, v),
            ....:                                     'elliptic paraboloid')
            sage: eparaboloid.natural_frame()
            {1: (1, 0, 2*u), 2: (0, 1, 2*v)}
        """

        dr1 = \
            vector([_simplify_full_rad( diff(f,self.variables[1]) )
                    for f in self.equation])
        dr2 = \
            vector([_simplify_full_rad( diff(f,self.variables[2]) )
                    for f in self.equation])

        return {1:dr1, 2:dr2}

    @cached_method
    def normal_vector(self, normalized=False):
        """
        Return the normal vector field of the parametrized surface.

        INPUT:

          - ``normalized`` -- boolean (default: ``False``); specifies whether
            the normal vector should be normalized

        OUTPUT: normal vector field

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: eparaboloid = ParametrizedSurface3D((u, v, u^2 + v^2), (u, v),
            ....:                                     'elliptic paraboloid')
            sage: eparaboloid.normal_vector(normalized=False)
            (-2*u, -2*v, 1)
            sage: eparaboloid.normal_vector(normalized=True)
            (-2*u/sqrt(4*u^2 + 4*v^2 + 1),
             -2*v/sqrt(4*u^2 + 4*v^2 + 1),
             1/sqrt(4*u^2 + 4*v^2 + 1))
        """

        dr = self.natural_frame()
        normal = dr[1].cross_product(dr[2])

        if normalized:
            normal /= normal.norm()
        return _simplify_full_rad(normal)

    @cached_method
    def _compute_first_fundamental_form_coefficient(self, index):
        """
        Helper function to compute coefficients of the first fundamental form.

        Do not call this method directly; instead use
        ``first_fundamental_form_coefficient``.
        This method is cached, and expects its argument to be a list.

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: eparaboloid = ParametrizedSurface3D((u, v, u^2+v^2), (u, v))
            sage: eparaboloid._compute_first_fundamental_form_coefficient((1,2))
            4*u*v
        """
        dr = self.natural_frame()
        return _simplify_full_rad(dr[index[0]]*dr[index[1]])

    def first_fundamental_form_coefficient(self, index):
        r"""
        Compute a single component `g_{ij}` of the first fundamental form.  If
        the parametric representation of the surface is given by the vector
        function `\vec r(u^i)`, where `u^i`, `i = 1, 2` are curvilinear
        coordinates, then `g_{ij} = \frac{\partial \vec r}{\partial u^i} \cdot \frac{\partial \vec r}{\partial u^j}`.

        INPUT:

        - ``index`` -- tuple ``(i, j)`` specifying the index of the component `g_{ij}`

        OUTPUT: component `g_{ij}` of the first fundamental form

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: eparaboloid = ParametrizedSurface3D((u, v, u^2+v^2), (u, v))
            sage: eparaboloid.first_fundamental_form_coefficient((1,2))
            4*u*v

        When the index is invalid, an error is raised::

            sage: u, v = var('u, v', domain='real')
            sage: eparaboloid = ParametrizedSurface3D((u, v, u^2+v^2), (u, v))
            sage: eparaboloid.first_fundamental_form_coefficient((1,5))
            Traceback (most recent call last):
            ...
            ValueError: Index (1, 5) out of bounds.
        """
        index = tuple(sorted(index))
        if len(index) == 2 and all(i == 1 or i == 2 for i in index):
            return self._compute_first_fundamental_form_coefficient(index)
        raise ValueError("Index %s out of bounds." % str(index))

    def first_fundamental_form_coefficients(self):
        r"""
        Return the coefficients of the first fundamental form as a dictionary.
        The keys are tuples `(i, j)`, where `i` and `j` range over `1, 2`,
        while the values are the corresponding coefficients `g_{ij}`.

        OUTPUT: dictionary of first fundamental form coefficients

        EXAMPLES::

            sage: u, v = var('u,v', domain='real')
            sage: sphere = ParametrizedSurface3D((cos(u)*cos(v), sin(u)*cos(v), sin(v)), (u, v), 'sphere')
            sage: sphere.first_fundamental_form_coefficients()
            {(1, 1): cos(v)^2, (1, 2): 0, (2, 1): 0, (2, 2): 1}
        """
        coefficients = {}
        for index in product((1, 2), repeat=2):
            coefficients[index] = \
                self._compute_first_fundamental_form_coefficient(index)
        return coefficients

    def first_fundamental_form(self, vector1, vector2):
        r"""
        Evaluate the first fundamental form on two vectors expressed with
        respect to the natural coordinate frame on the surface. In other words,
        if the vectors are `v = (v^1, v^2)` and `w = (w^1, w^2)`, calculate
        `g_{11} v^1 w^1 + g_{12}(v^1 w^2 + v^2 w^1) + g_{22} v^2 w^2`, with
        `g_{ij}` the coefficients of the first fundamental form.

        INPUT:

        - ``vector1``, ``vector2`` -- vectors on the surface

        OUTPUT: first fundamental form evaluated on the input vectors

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: v1, v2, w1, w2 = var('v1, v2, w1, w2', domain='real')
            sage: sphere = ParametrizedSurface3D((cos(u)*cos(v), sin(u)*cos(v), sin(v)), (u, v),'sphere')
            sage: sphere.first_fundamental_form(vector([v1,v2]),vector([w1,w2]))
            v1*w1*cos(v)^2 + v2*w2

            sage: vv = vector([1,2])
            sage: sphere.first_fundamental_form(vv,vv)
            cos(v)^2 + 4

            sage: sphere.first_fundamental_form([1,1],[2,1])
            2*cos(v)^2 + 1
        """
        gamma = self.first_fundamental_form_coefficients()
        return sum(gamma[(i,j)] * vector1[i - 1] * vector2[j - 1]
                   for i, j in product((1, 2), repeat=2))

    def area_form_squared(self):
        """
        Return the square of the coefficient of the area form on the surface.
        In terms of the coefficients `g_{ij}` (where `i, j = 1, 2`) of the
        first fundamental form, this invariant is given by
        `A^2 = g_{11}g_{22} - g_{12}^2`.

        See also :meth:`.area_form`.

        OUTPUT: square of the area form

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.area_form_squared()
            cos(v)^2
        """
        gamma = self.first_fundamental_form_coefficients()
        sq = gamma[(1,1)] * gamma[(2,2)] - gamma[(1,2)]**2
        return _simplify_full_rad(sq)

    def area_form(self):
        r"""
        Return the coefficient of the area form on the surface.  In terms of
        the coefficients `g_{ij}` (where `i, j = 1, 2`) of the first
        fundamental form, the coefficient of the area form is given by
        `A = \sqrt{g_{11}g_{22} - g_{12}^2}`.

        See also :meth:`.area_form_squared`.

        OUTPUT: coefficient of the area form

        EXAMPLES::

            sage: u, v = var('u,v', domain='real')
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.area_form()
            cos(v)
        """
        f = abs(sqrt(self.area_form_squared()))
        return _simplify_full_rad(f)

    def first_fundamental_form_inverse_coefficients(self):
        r"""
        Return the coefficients `g^{ij}` of the inverse of the fundamental
        form, as a dictionary.  The inverse coefficients are defined by
        `g^{ij} g_{jk} = \delta^i_k`  with `\delta^i_k` the Kronecker
        delta.

        OUTPUT: dictionary of the inverse coefficients

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.first_fundamental_form_inverse_coefficients()
            {(1, 1): cos(v)^(-2), (1, 2): 0, (2, 1): 0, (2, 2): 1}
        """

        g = self.first_fundamental_form_coefficients()
        D = g[(1,1)] * g[(2,2)] - g[(1,2)]**2

        gi11 = _simplify_full_rad(g[(2,2)]/D)
        gi12 = _simplify_full_rad(-g[(1,2)]/D)
        gi21 = gi12
        gi22 = _simplify_full_rad(g[(1,1)]/D)

        return {(1,1): gi11, (1,2): gi12, (2,1): gi21, (2,2): gi22}

    def first_fundamental_form_inverse_coefficient(self, index):
        r"""
        Return a specific component `g^{ij}` of the inverse of the fundamental
        form.

        INPUT:

        - ``index`` -- tuple ``(i, j)`` specifying the index of the component `g^{ij}`

        OUTPUT: component of the inverse of the fundamental form

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.first_fundamental_form_inverse_coefficient((1, 2))
            0
            sage: sphere.first_fundamental_form_inverse_coefficient((1, 1))
            cos(v)^(-2)
        """

        index = tuple(sorted(index))
        if len(index) == 2 and all(i == 1 or i == 2 for i in index):
            return self.first_fundamental_form_inverse_coefficients()[index]
        raise ValueError("Index %s out of bounds." % str(index))

    @cached_method
    def rotation(self, theta):
        r"""
        Give the matrix of the rotation operator over a given angle `\theta`
        with respect to the natural frame.

        INPUT:

        - ``theta`` -- rotation angle

        OUTPUT: rotation matrix with respect to the natural frame

        ALGORITHM:

        The operator of rotation over `\pi/2` is `J^i_j = g^{ik}\omega_{jk}`,
        where `\omega` is the area form.  The operator of rotation over an
        angle `\theta` is `\cos(\theta) I + \sin(\theta) J`.

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: assume(cos(v)>0)
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')

        We first compute the matrix of rotation over `\pi/3`::

            sage: rotation = sphere.rotation(pi/3); rotation
            [                1/2 -1/2*sqrt(3)/cos(v)]
            [ 1/2*sqrt(3)*cos(v)                 1/2]

        We verify that three successive rotations over `\pi/3` yield minus the identity::

            sage: rotation^3
            [-1  0]
            [ 0 -1]
        """

        from sage.functions.trig import sin, cos

        gi = self.first_fundamental_form_inverse_coefficients()
        w12 = self.area_form()
        R11 = (cos(theta) + sin(theta)*gi[1,2]*w12).simplify_full()
        R12 = (- sin(theta)*gi[1,1]*w12).simplify_full()
        R21 = (sin(theta)*gi[2,2]*w12).simplify_full()
        R22 = (cos(theta) - sin(theta)*gi[2,1]*w12).simplify_full()
        return matrix([[R11,R12],[R21,R22]])

    @cached_method
    def orthonormal_frame(self, coordinates='ext'):
        r"""
        Return the orthonormal frame field on the surface, expressed either
        in exterior coordinates (i.e. expressed as vector fields in the
        ambient space `\mathbb{R}^3`, the default) or interior coordinates
        (with respect to the natural frame)

        INPUT:

        - ``coordinates`` -- either ``ext`` (default) or ``int``

        OUTPUT: orthogonal frame field as a dictionary

        ALGORITHM:

        We normalize the first vector `\vec e_1` of the natural frame and then
        get the second frame vector as `\vec e_2 = [\vec n, \vec e_1]`, where
        `\vec n` is the unit normal to the surface.

        EXAMPLES::

            sage: u, v = var('u,v', domain='real')
            sage: assume(cos(v)>0)
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v), sin(u)*cos(v), sin(v)], [u, v],'sphere')
            sage: frame = sphere.orthonormal_frame(); frame
            {1: (-sin(u), cos(u), 0), 2: (-cos(u)*sin(v), -sin(u)*sin(v), cos(v))}
            sage: (frame[1]*frame[1]).simplify_full()
            1
            sage: (frame[1]*frame[2]).simplify_full()
            0
            sage: frame[1] == sphere.orthonormal_frame_vector(1)
            True

        We compute the orthonormal frame with respect to the natural frame on
        the surface::

            sage: frame_int = sphere.orthonormal_frame(coordinates='int'); frame_int
            {1: (1/cos(v), 0), 2: (0, 1)}
            sage: sphere.first_fundamental_form(frame_int[1], frame_int[1])
            1
            sage: sphere.first_fundamental_form(frame_int[1], frame_int[2])
            0
            sage: sphere.first_fundamental_form(frame_int[2], frame_int[2])
            1
        """
        from sage.symbolic.constants import pi

        if coordinates not in ['ext', 'int']:
            raise ValueError("Coordinate system must be exterior ('ext') "
                             "or interior ('int').")

        c = self.first_fundamental_form_coefficient([1, 1])
        if coordinates == 'ext':
            f1 = self.natural_frame()[1]

            E1 = _simplify_full_rad(f1 / sqrt(c))
            E2 = _simplify_full_rad(
                self.normal_vector(normalized=True).cross_product(E1))
        else:
            E1 = vector([_simplify_full_rad(1 / sqrt(c)), 0])
            E2 = (self.rotation(pi / 2) * E1).simplify_full()
        return {1: E1, 2: E2}

    def orthonormal_frame_vector(self, index, coordinates='ext'):
        r"""
        Return a specific basis vector field of the orthonormal frame field on
        the surface, expressed in exterior or interior coordinates.  See
        :meth:`orthogonal_frame` for more details.

        INPUT:

        - ``index`` -- index of the basis vector
        - ``coordinates`` -- either ``ext`` (default) or ``int``

        OUTPUT: orthonormal frame vector field

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: assume(cos(v)>0)
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: V1 = sphere.orthonormal_frame_vector(1); V1
            (-sin(u), cos(u), 0)
            sage: V2 = sphere.orthonormal_frame_vector(2); V2
            (-cos(u)*sin(v), -sin(u)*sin(v), cos(v))
            sage: (V1*V1).simplify_full()
            1
            sage: (V1*V2).simplify_full()
            0

            sage: n = sphere.normal_vector(normalized=True)
            sage: (V1.cross_product(V2) - n).simplify_full()
            (0, 0, 0)
        """

        return self.orthonormal_frame(coordinates)[index]

    def lie_bracket(self, v, w):
        r"""
        Return the Lie bracket of two vector fields that are tangent
        to the surface. The vector fields should be given in intrinsic
        coordinates, i.e. with respect to the natural frame.

        INPUT:

        - ``v``, ``w`` -- vector fields on the surface, expressed
          as pairs of functions or as vectors of length 2

        OUTPUT: the Lie bracket `[v, w]`

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: assume(cos(v)>0)
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.lie_bracket([u,v],[-v,u])
            (0, 0)

            sage: EE_int = sphere.orthonormal_frame(coordinates='int')
            sage: sphere.lie_bracket(EE_int[1],EE_int[2])
            (sin(v)/cos(v)^2, 0)
        """
        v = vector(SR, v)
        w = vector(SR, w)

        variables = self.variables_list
        Dv = matrix([[_simplify_full_rad(diff(component, u))
                      for u in variables] for component in v])
        Dw = matrix([[_simplify_full_rad(diff(component, u))
                      for u in variables] for component in w])
        return vector(Dv*w - Dw*v).simplify_full()

    def frame_structure_functions(self, e1, e2):
        r"""
        Return the structure functions `c^k_{ij}` for a frame field
        `e_1, e_2`, i.e. a pair of vector fields on the surface which are
        linearly independent at each point.  The structure functions are
        defined using the Lie bracket by `[e_i,e_j] = c^k_{ij}e_k`.

        INPUT:

        - ``e1``, ``e2`` -- vector fields in intrinsic coordinates on
          the surface, expressed as pairs of functions, or as vectors of
          length 2

        OUTPUT:

        Dictionary of structure functions, where the key ``(i, j, k)`` refers to
        the structure function `c_{i,j}^k`.

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: assume(cos(v) > 0)
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v), sin(u)*cos(v), sin(v)], [u, v], 'sphere')
            sage: sphere.frame_structure_functions([u, v], [-v, u])
            {(1, 1, 1): 0,
             (1, 1, 2): 0,
             (1, 2, 1): 0,
             (1, 2, 2): 0,
             (2, 1, 1): 0,
             (2, 1, 2): 0,
             (2, 2, 1): 0,
             (2, 2, 2): 0}

        We construct the structure functions of the orthonormal frame on the
        surface::

            sage: EE_int = sphere.orthonormal_frame(coordinates='int')
            sage: CC = sphere.frame_structure_functions(EE_int[1],EE_int[2]); CC
            {(1, 1, 1): 0,
             (1, 1, 2): 0,
             (1, 2, 1): sin(v)/cos(v),
             (1, 2, 2): 0,
             (2, 1, 1): -sin(v)/cos(v),
             (2, 1, 2): 0,
             (2, 2, 1): 0,
             (2, 2, 2): 0}
            sage: sphere.lie_bracket(EE_int[1],EE_int[2]) - CC[(1,2,1)]*EE_int[1] - CC[(1,2,2)]*EE_int[2]
            (0, 0)
        """
        e1 = vector(SR, e1)
        e2 = vector(SR, e2)

        lie_bracket = self.lie_bracket(e1, e2).simplify_full()
        transformation = matrix(SR, [e1, e2]).transpose()

        w = (transformation.inverse()*lie_bracket).simplify_full()

        return {(1,1,1): 0, (1,1,2): 0, (1,2,1): w[0], (1,2,2): w[1],
                (2,1,1): -w[0], (2,1,2): -w[1], (2,2,1): 0, (2,2,2): 0}

    @cached_method
    def _compute_second_order_frame_element(self, index):
        """
        Compute an element of the second order frame of the surface.  See
        :meth:`second_order_natural_frame` for more details.

        This method expects its arguments in tuple form for caching.
        As it does no input checking, it should not be called directly.

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: paraboloid = ParametrizedSurface3D([u, v, u^2 + v^2], [u,v], 'paraboloid')
            sage: paraboloid._compute_second_order_frame_element((1, 2))
            (0, 0, 0)
            sage: paraboloid._compute_second_order_frame_element((2, 2))
            (0, 0, 2)
        """
        variables = [self.variables[i] for i in index]
        ddr_element = vector([_simplify_full_rad(diff(f, variables))
                              for f in self.equation])

        return ddr_element

    def second_order_natural_frame(self):
        r"""
        Return the second-order frame of the surface, i.e. computes the
        second-order derivatives (with respect to the parameters on the
        surface) of the parametric expression `\vec r = \vec r(u^1,u^2)`
        of the surface.

        OUTPUT:

         - Dictionary where the keys are 2-tuples ``(i, j)`` and the values are the corresponding derivatives `r_{ij}`.

        EXAMPLES:

        We compute the second-order natural frame of the sphere::

            sage: u, v = var('u, v', domain='real')
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.second_order_natural_frame()
            {(1, 1): (-cos(u)*cos(v), -cos(v)*sin(u), 0),
             (1, 2): (sin(u)*sin(v), -cos(u)*sin(v), 0),
             (2, 1): (sin(u)*sin(v), -cos(u)*sin(v), 0),
             (2, 2): (-cos(u)*cos(v), -cos(v)*sin(u), -sin(v))}
        """

        vectors = {}
        for index in product((1, 2), repeat=2):
            sorted_index = tuple(sorted(index))
            vectors[index] = \
                self._compute_second_order_frame_element(sorted_index)
        return vectors

    def second_order_natural_frame_element(self, index):
        r"""
        Return a vector in the second-order frame of the surface, i.e.
        computes the second-order derivatives of the parametric expression
        `\vec{r}` of the surface with respect to the parameters listed in the
        argument.

        INPUT:

        - ``index`` -- a 2-tuple ``(i, j)`` specifying the element of the second-order frame

        OUTPUT: the second-order derivative `r_{ij}`

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.second_order_natural_frame_element((1, 2))
            (sin(u)*sin(v), -cos(u)*sin(v), 0)
        """

        index = tuple(sorted(index))
        if len(index) == 2 and all(i == 1 or i == 2 for i in index):
            return self._compute_second_order_frame_element(index)
        raise ValueError("Index %s out of bounds." % str(index))

    @cached_method
    def _compute_second_fundamental_form_coefficient(self, index):
        """
        Compute a coefficient of the second fundamental form of the surface.
        See ``second_fundamental_form_coefficient`` for more details.

        This method expects its arguments in tuple form for caching.  As it
        does no input checking, it should not be called directly.

        EXAMPLES::

            sage: u, v = var('u,v', domain='real')
            sage: paraboloid = ParametrizedSurface3D([u, v, u^2+v^2], [u, v], 'paraboloid')
            sage: paraboloid._compute_second_fundamental_form_coefficient((1,1))
            2/sqrt(4*u^2 + 4*v^2 + 1)
        """
        N = self.normal_vector(normalized=True)
        v = self.second_order_natural_frame_element(index)
        return _simplify_full_rad(v*N)

    def second_fundamental_form_coefficient(self, index):
        r"""
        Return the coefficient `h_{ij}` of the second fundamental form
        corresponding to the index `(i, j)`.  If the equation of the surface
        is `\vec{r}(u^1, u^2)`, then `h_{ij} = \vec{r}_{u^i u^j} \cdot \vec{n}`,
        where `\vec{n}` is the unit normal.

        INPUT:

        - ``index`` -- a 2-tuple ``(i, j)``

        OUTPUT: component `h_{ij}` of the second fundamental form

        EXAMPLES::

            sage: u, v = var('u,v', domain='real')
            sage: assume(cos(v)>0)
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.second_fundamental_form_coefficient((1, 1))
            -cos(v)^2
            sage: sphere.second_fundamental_form_coefficient((2, 1))
            0
        """
        index = tuple(index)
        if len(index) == 2 and all(i == 1 or i == 2 for i in index):
            return self._compute_second_fundamental_form_coefficient(index)
        raise ValueError("Index %s out of bounds." % str(index))

    def second_fundamental_form_coefficients(self):
        """
        Return the coefficients `h_{ij}` of the second fundamental form as
        a dictionary, where the keys are the indices `(i, j)` and the values
        are the corresponding components `h_{ij}`.

        When only one component is needed, consider instead the function
        :meth:`second_fundamental_form_coefficient`.

        OUTPUT: dictionary of second fundamental form coefficients

        EXAMPLES::

            sage: u, v = var('u, v', domain='real')
            sage: assume(cos(v)>0)
            sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
            sage: sphere.second_fundamental_form_coefficients()
            {(1, 1): -cos(v)^2, (1, 2): 0, (2, 1): 0, (2, 2): -1}
        """

        coefficients = {}
        for index in product((1, 2), repeat=2):
            coefficients[index] = \
                self._compute_second_fundamental_form_coefficient(index)
        return coefficients

    def second_fundamental_form(self, vector1, vector2):
        r"""
        Evaluates the second fundamental form on two vectors on the surface.
        If the vectors are given by `v=(v^1,v^2)` and `w=(w^1,w^2)`, the
        result of this function is `h_{11} v^1 w^1 + h_{12}(v^1 w^2 + v^2 w^1) + h_{22} v^2 w^2`.

        INPUT:

        - ``vector1``, ``vector2`` -- 2-tuples representing the input vectors

        OUTPUT: value of the second fundamental form evaluated on the given vectors

        EXAMPLES:

        We evaluate the second fundamental form on two symbolic vectors::

           sage: u, v = var('u, v', domain='real')
           sage: v1, v2, w1, w2 = var('v1, v2, w1, w2', domain='real')
           sage: assume(cos(v) > 0)
           sage: sphere = ParametrizedSurface3D([cos(u)*cos(v),sin(u)*cos(v),sin(v)],[u,v],'sphere')
           sage: sphere.second_fundamental_form(vector([v1, v2]), vector([w1, w2]))
           -v1*w1*cos(v)^2 - v2*w2

        We evaluate the second fundamental form on vectors with numerical
        components::

           sage: vect = vector([1,2])
           sage: sphere.second_fundamental_form(vect, vect)
           -cos(v)^2 - 4
           sage: sphere.second_fundamental_form([1,1], [2,1])
           -2*cos(v)^2 - 1
        """
        hh = self.second_fundamental_form_coefficients()
        return sum(hh[(i, j)] * vector1[i - 1] * vector2[j - 1]
                   for (i, j) in product((1, 2), repeat=2))

    def gauss_curvature(self):
        r"""
        Finds the gaussian curvature of the surface, given by
        `K = \frac{h_{11}h_{22} - h_{12}^2}{g_{11}g_{22} - g_{12}^2}`,
        where `g_{ij}` and `h_{ij}` are the coefficients of the first
        and second fundamental form, respectively.

        OUTPUT: Gaussian curvature of the surface.

        EXAMPLES::

           sage: R = var('R')
           sage: assume(R>0)
           sage: u, v = var('u,v', domain='real')
           sage: assume(cos(v)>0)
           sage: sphere = ParametrizedSurface3D([R*cos(u)*cos(v),R*sin(u)*cos(v),R*sin(v)],[u,v],'sphere')
           sage: sphere.gauss_curvature()
           R^(-2)
        """
        hh = self.second_fundamental_form_coefficients()
        return _simplify_full_rad(
            (hh[(1,1)] * hh[(2,2)] - hh[(1,2)]**2)/self.area_form_squared())

    def mean_curvature(self):
        r"""
        Finds the mean curvature of the surface, given by
        `H = \frac{1}{2}\frac{g_{22}h_{11} - 2g_{12}h_{12} + g_{11}h_{22}}{g_{11}g_{22} - g_{12}^2}`,
        where `g_{ij}` and `h_{ij}` are the components of the first and second
        fundamental forms, respectively.

        OUTPUT: mean curvature of the surface

        EXAMPLES::

           sage: R = var('R')
           sage: assume(R>0)
           sage: u, v = var('u,v', domain='real')
           sage: assume(cos(v)>0)
           sage: sphere = ParametrizedSurface3D([R*cos(u)*cos(v),R*sin(u)*cos(v),R*sin(v)],[u,v],'sphere')
           sage: sphere.mean_curvature()
           -1/R
        """
        gg = self.first_fundamental_form_coefficients()
        hh = self.second_fundamental_form_coefficients()
        denom = 2*self.area_form_squared()
        numer = (gg[(2,2)]*hh[(1,1)] - 2*gg[(1,2)]*hh[(1,2)] +
                 gg[(1,1)]*hh[(2,2)]).simplify_full()
        return _simplify_full_rad(numer/denom)

    @cached_method
    def shape_operator_coefficients(self):
        r"""
        Return the components of the shape operator of the surface as a
        dictionary. See ``shape_operator`` for more information.

        OUTPUT:

        Dictionary where the keys are two-tuples ``(i, j)``, with values the
        corresponding component of the shape operator.

        EXAMPLES::

           sage: R = var('R')
           sage: u, v = var('u,v', domain='real')
           sage: assume(cos(v)>0)
           sage: sphere = ParametrizedSurface3D([R*cos(u)*cos(v),R*sin(u)*cos(v),R*sin(v)],[u,v],'sphere')
           sage: sphere.shape_operator_coefficients()
           {(1, 1): -1/R, (1, 2): 0, (2, 1): 0, (2, 2): -1/R}
        """

        gi = self.first_fundamental_form_inverse_coefficients()
        hh = self.second_fundamental_form_coefficients()

        sh_op11 = _simplify_full_rad(gi[(1,1)]*hh[(1,1)] + gi[(1,2)]*hh[(1,2)])
        sh_op12 = _simplify_full_rad(gi[(1,1)]*hh[(2,1)] + gi[(1,2)]*hh[(2,2)])
        sh_op21 = _simplify_full_rad(gi[(2,1)]*hh[(1,1)] + gi[(2,2)]*hh[(1,2)])
        sh_op22 = _simplify_full_rad(gi[(2,1)]*hh[(2,1)] + gi[(2,2)]*hh[(2,2)])

        return {(1,1): sh_op11, (1,2): sh_op12, (2,1): sh_op21, (2,2): sh_op22}

    def shape_operator(self):
        r"""
        Return the shape operator of the surface as a matrix.  The shape
        operator is defined as the derivative of the Gauss map, and is
        computed here in terms of the first and second fundamental form by
        means of the Weingarten equations.

        OUTPUT: matrix of the shape operator

        EXAMPLES::

           sage: R = var('R')
           sage: assume(R>0)
           sage: u, v = var('u,v', domain='real')
           sage: assume(cos(v)>0)
           sage: sphere = ParametrizedSurface3D([R*cos(u)*cos(v),R*sin(u)*cos(v),R*sin(v)],[u,v],'sphere')
           sage: S = sphere.shape_operator(); S
           [-1/R    0]
           [   0 -1/R]

        The eigenvalues of the shape operator are the principal curvatures of
        the surface::

            sage: u, v = var('u,v', domain='real')
            sage: paraboloid = ParametrizedSurface3D([u, v, u^2+v^2], [u, v], 'paraboloid')
            sage: S = paraboloid.shape_operator(); S
            [2*(4*v^2 + 1)/(4*u^2 + 4*v^2 + 1)^(3/2)        -8*u*v/(4*u^2 + 4*v^2 + 1)^(3/2)]
            [       -8*u*v/(4*u^2 + 4*v^2 + 1)^(3/2) 2*(4*u^2 + 1)/(4*u^2 + 4*v^2 + 1)^(3/2)]
            sage: S.eigenvalues()
            [2*sqrt(4*u^2 + 4*v^2 + 1)/(16*u^4 + 16*v^4 + 8*(4*u^2 + 1)*v^2 + 8*u^2 + 1), 2/sqrt(4*u^2 + 4*v^2 + 1)]
        """

        shop = self.shape_operator_coefficients()
        shop_matrix = matrix([[shop[(1,1)],shop[(1,2)]],
                            [shop[(2,1)],shop[(2,2)]]])
        return shop_matrix

    def principal_directions(self):
        r"""
        Finds the principal curvatures and principal directions of the surface.

        OUTPUT:

        For each principal curvature, returns a list of the form
        `(\rho, V, n)`, where `\rho` is the principal curvature,
        `V` is the corresponding principal direction, and `n` is
        the multiplicity.

        EXAMPLES::

           sage: u, v = var('u, v', domain='real')
           sage: R, r = var('R,r', domain='real')
           sage: assume(R>r,r>0)
           sage: torus = ParametrizedSurface3D([(R+r*cos(v))*cos(u),(R+r*cos(v))*sin(u),r*sin(v)],[u,v],'torus')
           sage: torus.principal_directions()
           [(-cos(v)/(r*cos(v) + R), [(1, 0)], 1), (-1/r, [(0, 1)], 1)]

        ::

            sage: u, v = var('u, v', domain='real')
            sage: V = vector([u*cos(u+v), u*sin(u+v), u+v])
            sage: helicoid = ParametrizedSurface3D(V, (u, v))
            sage: helicoid.principal_directions()
            [(-1/(u^2 + 1), [(1, -(u^2 - sqrt(u^2 + 1) + 1)/(u^2 + 1))], 1),
            (1/(u^2 + 1), [(1, -(u^2 + sqrt(u^2 + 1) + 1)/(u^2 + 1))], 1)]
        """
        return self.shape_operator().eigenvectors_right()

    @cached_method
    def connection_coefficients(self):
        r"""
        Compute the connection coefficients or Christoffel symbols
        `\Gamma^k_{ij}` of the surface. If the coefficients of the first
        fundamental form are given by `g_{ij}` (where `i, j = 1, 2`), then
        `\Gamma^k_{ij} = \frac{1}{2} g^{kl} \left( \frac{\partial g_{li}}{\partial x^j}
        - \frac{\partial g_{ij}}{\partial x^l}
        + \frac{\partial g_{lj}}{\partial x^i} \right)`.
        Here, `(g^{kl})` is the inverse of the matrix `(g_{ij})`, with
        `i, j = 1, 2`.

        OUTPUT:

        Dictionary of connection coefficients, where the keys are 3-tuples
        `(i,j,k)` and the values are the corresponding coefficients
        `\Gamma^k_{ij}`.

        EXAMPLES::

           sage: r = var('r')
           sage: assume(r > 0)
           sage: u, v = var('u,v', domain='real')
           sage: assume(cos(v)>0)
           sage: sphere = ParametrizedSurface3D([r*cos(u)*cos(v),r*sin(u)*cos(v),r*sin(v)],[u,v],'sphere')
           sage: sphere.connection_coefficients()
           {(1, 1, 1): 0,
            (1, 1, 2): cos(v)*sin(v),
            (1, 2, 1): -sin(v)/cos(v),
            (1, 2, 2): 0,
            (2, 1, 1): -sin(v)/cos(v),
            (2, 1, 2): 0,
            (2, 2, 1): 0,
            (2, 2, 2): 0}
        """
        x = self.variables
        gg = self.first_fundamental_form_coefficients()
        gi = self.first_fundamental_form_inverse_coefficients()

        dg = {}
        for i,j,k in product((1, 2), repeat=3):
            dg[(i,j,k)] = _simplify_full_rad(gg[(j,k)].differentiate(x[i]))

        structfun = {}
        for i,j,k in product((1, 2), repeat=3):
            structfun[(i,j,k)] = sum(gi[(k,s)]*(dg[(i,j,s)] + dg[(j,i,s)]
                                                - dg[(s,i,j)])/2
                                     for s in (1,2))
            structfun[(i,j,k)] = _simplify_full_rad(structfun[(i,j,k)])
        return structfun

    @cached_method
    def _create_geodesic_ode_system(self):
        r"""
        Helper method to create a fast floating-point version of the
        geodesic equations, used by :meth:`geodesics_numerical`.

        EXAMPLES::

           sage: p, q = var('p,q', domain='real')
           sage: sphere = ParametrizedSurface3D([cos(q)*cos(p),sin(q)*cos(p),sin(p)],[p,q],'sphere')
           sage: ode = sphere._create_geodesic_ode_system()
           sage: ode.function(0.0, (1.0, 0.0, 1.0, 1.0))
           [1.00000000000000, 1.00000000000000, -0.4546487134128409, 3.114815449309804]
        """
        from sage.ext.fast_eval import fast_float
        from sage.calculus.ode import ode_solver

        u1 = self.variables[1]
        u2 = self.variables[2]

        C = self.connection_coefficients()

        with SR.temp_var(domain='real') as v1:
            with SR.temp_var(domain='real') as v2:
                dv1 = - C[(1,1,1)]*v1**2 - 2*C[(1,2,1)]*v1*v2 - C[(2,2,1)]*v2**2
                dv2 = - C[(1,1,2)]*v1**2 - 2*C[(1,2,2)]*v1*v2 - C[(2,2,2)]*v2**2
                fun1 = fast_float(dv1, str(u1), str(u2), str(v1), str(v2))
                fun2 = fast_float(dv2, str(u1), str(u2), str(v1), str(v2))

                geodesic_ode = ode_solver()
                geodesic_ode.function = (
                    lambda t, u1_u2_v1_v2:
                    [u1_u2_v1_v2[2], u1_u2_v1_v2[3], fun1(*u1_u2_v1_v2), fun2(*u1_u2_v1_v2)])
                return geodesic_ode

    def geodesics_numerical(self, p0, v0, tinterval):
        r"""
        Numerical integration of the geodesic equations.  Explicitly, the
        geodesic equations are given by
        `\frac{d^2 u^i}{dt^2} + \Gamma^i_{jk} \frac{d u^j}{dt} \frac{d u^k}{dt} = 0`.

        Solving these equations gives the coordinates `(u^1, u^2)` of
        the geodesic on the surface.  The coordinates in space can
        then be found by substituting `(u^1, u^2)` into the vector
        `\vec{r}(u^1, u^2)` representing the surface.

        ALGORITHM:

        The geodesic equations are integrated forward in time using
        the ode solvers from ``sage.calculus.ode``.  See the member
        function ``_create_geodesic_ode_system`` for more details.

        INPUT:

        - ``p0`` -- 2-tuple with coordinates of the initial point

        - ``v0`` -- 2-tuple with components of the initial tangent vector to the geodesic

        - ``tinterval`` -- list ``[a, b, M]``, where ``(a,b)`` is the domain
          of the geodesic and ``M`` is the number of subdivision points used
          when returning the solution

        OUTPUT:

        List of lists ``[t, [u1(t), u2(t)], [v1(t), v2(t)], [x1(t), x2(t), x3(t)]]``, where

        - ``t`` -- a subdivision point;

        - ``[u1(t), u2(t)]`` are the intrinsic coordinates of the geodesic point;

        - ``[v1(t), v2(t)]`` are the intrinsic coordinates of the tangent vector to the geodesic;

        - ``[x1(t), x2(t), x3(t)]`` are the coordinates of the geodesic point in the three-dimensional space.

        EXAMPLES::

           sage: p, q = var('p,q', domain='real')
           sage: assume(cos(q)>0)
           sage: sphere = ParametrizedSurface3D([cos(q)*cos(p),sin(q)*cos(p),sin(p)],[p,q],'sphere')
           sage: geodesic = sphere.geodesics_numerical([0.0,0.0],[1.0,1.0],[0,2*pi,5])
           sage: times, points, tangent_vectors, ext_points = zip(*geodesic)

           sage: round4 = lambda vec: [N(x, digits=4) for x in vec]   # helper function to round to 4 digits
           sage: round4(times)
           [0.0000, 1.257, 2.513, 3.770, 5.027, 6.283]
           sage: [round4(p) for p in points]
           [[0.0000, 0.0000], [0.7644, 1.859], [-0.2876, 3.442], [-0.6137, 5.502], [0.5464, 6.937], [0.3714, 9.025]]
           sage: [round4(p) for p in ext_points]
           [[1.000, 0.0000, 0.0000], [-0.2049, 0.6921, 0.6921], [-0.9160, -0.2836, -0.2836], [0.5803, -0.5759, -0.5759], [0.6782, 0.5196, 0.5196], [-0.8582, 0.3629, 0.3629]]
        """
        solver = self._create_geodesic_ode_system()

        t_interval, n = tinterval[0:2], tinterval[2]
        solver.y_0 = [p0[0], p0[1], v0[0], v0[1]]
        solver.ode_solve(t_span=t_interval, num_points=n)

        parsed_solution = \
          [[vec[0], vec[1][0:2], vec[1][2:], self.point(vec[1])]
           for vec in solver.solution]

        return parsed_solution

    @cached_method
    def _create_pt_ode_system(self, curve, t):
        """
        Helper method to create a fast floating-point version of the parallel
        transport equations, used by ``parallel_translation_numerical``.

        INPUT:

        - ``curve`` -- curve in intrinsic coordinates along which to do parallel transport
        - ``t`` -- curve parameter

        EXAMPLES::

           sage: p, q = var('p,q', domain='real')
           sage: sphere = ParametrizedSurface3D([cos(q)*cos(p),sin(q)*cos(p),sin(p)],[p,q],'sphere')
           sage: s = var('s')
           sage: ode = sphere._create_pt_ode_system((s, s), s)
           sage: ode.function(0.0, (1.0, 1.0))
           [-0.0, 0.0]
        """

        from sage.ext.fast_eval import fast_float
        from sage.calculus.ode import ode_solver

        u1 = self.variables[1]
        u2 = self.variables[2]

        du1 = diff(curve[0], t)
        du2 = diff(curve[1], t)

        C = self.connection_coefficients()
        for coef in C:
            C[coef] = C[coef].subs({u1: curve[0], u2: curve[1]})

        with SR.temp_var(domain='real') as v1:
            with SR.temp_var(domain='real') as v2:
                dv1 = - C[(1,1,1)]*v1*du1 - C[(1,2,1)]*(du1*v2 + du2*v1) - \
                    C[(2,2,1)]*du2*v2
                dv2 = - C[(1,1,2)]*v1*du1 - C[(1,2,2)]*(du1*v2 + du2*v1) - \
                    C[(2,2,2)]*du2*v2
                fun1 = fast_float(dv1, str(t), str(v1), str(v2))
                fun2 = fast_float(dv2, str(t), str(v1), str(v2))

                pt_ode = ode_solver()
                pt_ode.function = lambda t, v1_v2: [fun1(t, v1_v2[0], v1_v2[1]), fun2(t, v1_v2[0], v1_v2[1])]
                return pt_ode

    def parallel_translation_numerical(self, curve, t, v0, tinterval):
        r"""
        Numerically solve the equations for parallel translation of a vector
        along a curve on the surface.  Explicitly, the equations for parallel
        translation are given by
        `\frac{d u^i}{dt} + u^j \frac{d c^k}{dt} \Gamma^i_{jk} = 0`,
        where `\Gamma^i_{jk}` are the connection coefficients of the surface,
        the vector to be transported has components `u^j` and the curve along
        which to transport has components `c^k`.

        ALGORITHM:

        The parallel transport equations are integrated forward in time using
        the ode solvers from ``sage.calculus.ode``. See :meth:`_create_pt_ode_system`
        for more details.

        INPUT:

        - ``curve`` -- 2-tuple of functions which determine the curve with respect to
          the local coordinate system

        - ``t`` -- symbolic variable denoting the curve parameter

        - ``v0`` -- 2-tuple representing the initial vector

        - ``tinterval`` -- list ``[a, b, N]``, where ``(a, b)`` is the domain of the curve
          and ``N`` is the number of subdivision points

        OUTPUT:

        The list consisting of lists ``[t, [v1(t), v2(t)]]``, where

        - ``t`` -- a subdivision point;

        - ``[v1(t), v2(t)]`` is the list of coordinates of the vector parallel translated
          along the curve.

        EXAMPLES::

           sage: p, q = var('p,q', domain='real')
           sage: v = [p,q]
           sage: assume(cos(q)>0)
           sage: sphere = ParametrizedSurface3D([cos(q)*cos(p),sin(q)*cos(p),sin(p)],v,'sphere')
           sage: s = var('s')
           sage: vector_field = sphere.parallel_translation_numerical([s,s],s,[1.0,1.0],[0.0, pi/4, 5])
           sage: times, components = zip(*vector_field)

           sage: round4 = lambda vec: [N(x, digits=4) for x in vec]   # helper function to round to 4 digits
           sage: round4(times)
           [0.0000, 0.1571, 0.3142, 0.4712, 0.6283, 0.7854]
           sage: [round4(v) for v in components]
           [[1.000, 1.000], [0.9876, 1.025], [0.9499, 1.102], [0.8853, 1.238], [0.7920, 1.448], [0.6687, 1.762]]
        """
        solver = self._create_pt_ode_system(tuple(curve), t)

        t_interval, n = tinterval[0:2], tinterval[2]
        solver.y_0 = v0
        solver.ode_solve(t_span=t_interval, num_points=n)

        return solver.solution
